import unittest
from frads import room


class TestRoom(unittest.TestCase):
    def test_make_room(self):
        width = 3
        depth = 5
        floor_floor = 4
        floor_ceiling = 3
        windows = [
            [1., 1., 1., 1.],
        ]
        thickness = 0.1
        aroom = room.create_south_facing_room(
            width=width,
            depth=depth,
            floor_floor=floor_floor,
            floor_ceiling=floor_ceiling,
            wpd=windows,
            swall_thickness=thickness,
        )

if __name__ == "__main__":
    unittest.main()



"""Typical Radiance matrix-based simulation workflows"""

import hashlib
import logging
import os
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from pathlib import Path
from shutil import rmtree

from frads.geom import Polygon, parse_polygon
from frads.matrix import (
    load_matrix,
    SensorSender,
    SurfaceSender,
    SurfaceReceiver,
    ViewSender,
    SunReceiver,
    SkyReceiver,
    load_binary_matrix,
    BASIS_DIMENSION,
    parse_rad_header,
    Matrix,
    matrix_multiply_rgb,
    SunMatrix,
    to_sparse_matrix3,
    sparse_matrix_multiply_rgb_vtds,
)
from frads.sky import parse_epw, parse_wea, WeaMetaData, WeaData, gen_perez_sky
from frads.utils import random_string
import numpy as np
import pyradiance as pr
from pyradiance import parse_view
from scipy.sparse import csr_matrix


logger = logging.getLogger("frads.methods")


@dataclass(slots=True)
class SceneConfig:
    """Radiance scene configuration object.

    It can be initialized with either a raw data bytes or a list
    of files. If a list of files is provided, they will be concatenated in
    the order they are provided.

    Attributes:
        files: A list of files to be concatenated to form the scene.
        for name, material in self.model.material.items():
            materials[name] = parse_material(name, material)
        for name, material in  self.model.material_no_mass.items():
            materials[name] = parse_material_no_mass(name, material)
        for name, material in self.model.window_material_simple_glazing_system.items():
            materials[name] = parse_window_material_simple_glazing_system(name, material)
        for name, material in self.model.window_material_glazing.items():
            materials[name] = parse_window_material_glazing(name, material)
        for name, material in self.model.window_material_blind.items():
            materials.update(parse_window_material_blind(material))
        bytes: A raw data string to be used as the scene.
        files_mtime: Files last modification time.
    """

    files: list[Path] = field(default_factory=list)
    bytes: bytes = b""
    files_mtime: list[float] = field(init=False, default_factory=list)

    def __post_init__(self):
        if len(self.files) > 0:
            for fpath in self.files:
                self.files_mtime.append(os.path.getmtime(fpath))


@dataclass(slots=True)
class MatrixConfig:
    matrix_file: str | Path = ""
    matrix_data: None | np.ndarray = None

    def __post_init__(self):
        if self.matrix_data is None:
            self.matrix_data = load_matrix(self.matrix_file)
        elif isinstance(self.matrix_data, list):
            self.matrix_data = np.array(self.matrix_data)


@dataclass(slots=True)
class ShadingGeometryConfig:
    shading_geometry_file: str | Path = ""
    shading_geometry_bytes: None | np.ndarray = None

    def __post_init__(self): ...


@dataclass(slots=True)
class MaterialConfig:
    """Material file/data configuration object.

    It can be initialized with either a raw data bytes or a list
    of files. If a list of files is provided, they will be concatenated in
    the order they are provided.

    Attributes:
        file: A file to be used as the material.
        bytes: A raw data string to be used as the material.
        matrices: A dictionary of matrix files/data.
        glazing_materials: A dictionary of glazing materials used for edgps calculations.
        file_mtime: File last modification time.

    Raises:
        ValueError: If no file, bytes, or matrices are provided.
    """

    files: list[Path] = field(default_factory=list)
    bytes: bytes = b""
    matrices: dict[str, MatrixConfig] = field(default_factory=dict)
    matrices_mlnp: dict[str, MatrixConfig] = field(default_factory=dict)
    glazing_materials: dict[str, pr.Primitive] = field(default_factory=dict)
    files_mtime: list[float] = field(init=False, default_factory=list)

    def __post_init__(self):
        if len(self.files) > 0:
            for fpath in self.files:
                self.files_mtime.append(os.path.getmtime(fpath))
        for k, v in self.matrices.items():
            if isinstance(v, dict):
                self.matrices[k] = MatrixConfig(**v)
        if self.bytes == b"" and len(self.files) == 0 and len(self.matrices) == 0:
            raise ValueError("MaterialConfig must have either file, bytes or matrices")


@dataclass(slots=True)
class WindowConfig:
    """Window file/data configuration object.

    Each WindowConfig instance coresponds to a window group, which
    can be initialized with either a file path or byte strings.
    In addition, the BSDF matrix files/data, high resolution tensor
    tree files, and shading geometry files or bytestring data can
    be initialized as well.

    Attributes:
        file: A file to be used as the window group.
        bytes: A raw data string to be used as the window group.
        matrix_name: A matrix name to be used for the window group.
        proxy_geometry: A raw data string to be used as the shading geometry.
        files_mtime: Files last modification time.

    Raises:
        ValueError: If neither file nor bytes are provided.
    """

    file: str | Path = ""
    bytes: bytes = b""
    matrix_name: str = ""
    polygon: None | Polygon = None
    proxy_geometry: dict[str, bytes] = field(default_factory=dict)
    files_mtime: list[float] = field(init=False, default_factory=list)

    def __post_init__(self):
        if os.path.exists(self.file):
            self.files_mtime.append(os.path.getmtime(self.file))
            if not isinstance(self.file, Path):
                self.file = Path(self.file)
        if self.bytes == b"":
            if self.file != "":
                with open(self.file, "rb") as f:
                    self.bytes = f.read()
            else:
                raise ValueError("WindowConfig must have either file or bytes")


@dataclass(slots=True)
class SensorConfig:
    """
    A configuration class for sensors that includes information on the file,
    data, and file modification time.

    Attributes:
        file: Path to the file containing sensor data. Default is an empty string.
        data: list of lists containing float data.
            Default is an empty list.
        file_mtime: Modification time of the file. This attribute is
            automatically initialized based on the 'file' attribute.

    Raises:
        ValueError: If neither file nor data are provided.
    """

    file: str = ""
    data: list[list[float]] = field(default_factory=list)
    file_mtime: float = field(init=False, default=0.0)

    def __post_init__(self):
        """
        Post-initialization method to set the file modification time and load data
        from the file if necessary.
        """
        if self.file != "":
            self.file_mtime = os.path.getmtime(self.file)
        if len(self.data) == 0:
            if self.file != "":
                with open(self.file) as f:
                    self.data = [
                        [float(i) for i in line.split()] for line in f.readlines()
                    ]
            else:
                raise ValueError("SensorConfig must have either file or data")


@dataclass(slots=True)
class ViewConfig:
    """
    A configuration class for views that includes information on the file,
    data, x/y resoluation, and file modification time.

    Attributes:
        file: Path to the file containing view data. Default is an empty string.
        view: A View object. Default is an empty string.
        xres: X resolution of the view. Default is 512.
        yres: Y resolution of the view. Default is 512.
        file_mtime: Modification time of the file. This attribute is
            automatically initialized based on the 'file' attribute.

    Raises:
        ValueError: If neither file nor view are provided.
    """

    file: str | Path = ""
    view: pr.View | str = field(default_factory=str)
    xres: int = 512
    yres: int = 512
    file_mtime: float = field(init=False, default=0.0)

    def __post_init__(self):
        if self.file != "":
            self.file_mtime = os.path.getmtime(self.file)
            if not isinstance(self.file, Path):
                self.file = Path(self.file)
        if os.path.exists(self.file) and self.view == "":
            self.view = pr.viewfile(str(self.file))
        elif self.view != "":
            if not isinstance(self.view, pr.View):
                self.view = parse_view(self.view)
        else:
            raise ValueError("ViewConfig must have either file or view")


@dataclass(slots=True)
class SurfaceConfig:
    """
    A configuration class for surfaces that includes information on the file,
    data, basis, and file modification time.

    Attributes:
        file: Path to the file containing surface data. Default is an empty string.
        primitives: A list of primitives. Default is an empty list.
        basis: A string representing the basis. Default is 'u'.
        file_mtime: Modification time of the file. This attribute is
            automatically initialized based on the 'file' attribute.

    Raises:
        ValueError: If neither file nor primitives are provided.
    """

    file: str | Path = ""
    primitives: list[pr.Primitive] = field(default_factory=list)
    basis: str = "u"
    file_mtime: float = field(init=False, default=0.0)

    def __post_init__(self):
        if self.file != "":
            self.file_mtime = os.path.getmtime(self.file)
        if not isinstance(self.file, Path):
            self.file = Path(self.file)
        if self.file.exists() and len(self.primitives) == 0:
            self.primitives = pr.parse_primitive(self.file.read_text())
        elif len(self.primitives) == 0:
            raise ValueError("SurfaceConfig must have either file or primitives")


@dataclass(slots=True)
class Settings:
    """Settings is a dataclass that holds the settings for a Radiance simulation.

    Attributes:
        name: The name of the simulation.
        num_processors: The number of processors to use for the simulation.
        method: The Radiance method to use for the simulation.
        overwrite: Whether to overwrite existing files.
        save_matrices: Whether to save the matrices generated by the simulation.
        sky_basis: The sky basis to use for the simulation.
        window_basis: The window basis to use for the simulation.
        non_coplanar_basis: The non-coplanar basis to use for the simulation.
        sun_basis: The sun basis to use for the simulation.
        sun_culling: Whether to cull suns.
        separate_direct: Whether to separate direct and indirect contributions.
        epw_file: The path to the EPW file to use for the simulation.
        wea_file: The path to the WEA file to use for the simulation.
        start_hour: The start hour for the simulation.
        end_hour: The end hour for the simulation.
        daylight_hours_only: Whether to simulate only daylight hours.
        latitude: The latitude for the simulation.
        longitude: The longitude for the simulation.
        timezone: The timezone for the simulation.
        orientation: sky rotation.
        site_elevation: The elevation for the simulation.
        sensor_sky_matrix: The sky matrix sampling parameters
        view_sky_matrix: View sky matrix sampling parameters
        sensor_sun_matrix: Sensor sun matrix sampling parameters
        view_sun_matrix: View sun matrix sampling parameters
        sensor_window_matrix: Sensor window matrix sampling parameters
        view_window_matrix: View window matrix sampling parameters
        daylight_matrix: Daylight matrix sampling parameters
    """

    name: str = field(default="")
    num_processors: int = 4
    method: str = field(default="3phase")
    overwrite: bool = False
    save_matrices: bool = False
    matrix_dir: str = field(default="Matrices")
    sky_basis: str = field(default="r1")
    window_basis: str = field(default="kf")
    non_coplanar_basis: str = field(default="kf")
    sun_basis: str = field(default="r6")
    sun_culling: bool = field(default=True)
    separate_direct: bool = field(default=False)
    epw_file: str = field(default="")
    wea_file: str = field(default="")
    start_hour: float = field(default=8)
    end_hour: float = field(default=18)
    daylight_hours_only: bool = True
    latitude: float = field(default=37)
    longitude: float = field(default=122)
    time_zone: int = field(default=120)
    orientation: float = field(default=0)
    site_elevation: float = field(default=100)
    sensor_sky_matrix: list[str] = field(
        default_factory=lambda: ["-ab", "6", "-ad", "8192", "-lw", "5e-5"]
    )
    sensor_sun_matrix: list[str] = field(
        default_factory=lambda: [
            "-ab",
            "1",
            "-ad",
            "256",
            "-lw",
            "1e-3",
            "-dj",
            "0",
            "-st",
            "0",
        ]
    )
    view_sun_matrix: list[str] = field(
        default_factory=lambda: [
            "-ab",
            "1",
            "-ad",
            "256",
            "-lw",
            "1e-3",
            "-dj",
            "0",
            "-st",
            "0",
        ]
    )
    view_sky_matrix: list[str] = field(
        default_factory=lambda: ["-ab", "6", "-ad", "8192", "-lw", "5e-5"]
    )
    daylight_matrix: list[str] = field(
        default_factory=lambda: ["-ab", "2", "-c", "5000"]
    )
    sensor_window_matrix: list[str] = field(
        default_factory=lambda: ["-ab", "5", "-ad", "8192", "-lw", "5e-5"]
    )
    surface_window_matrix: list[str] = field(
        default_factory=lambda: [
            "-ab",
            "5",
            "-ad",
            "8192",
            "-lw",
            "5e-5",
            "-c",
            "10000",
        ]
    )
    view_window_matrix: list[str] = field(
        default_factory=lambda: ["-ab", "5", "-ad", "8192", "-lw", "5e-5"]
    )
    files_mtime: list[float] = field(init=False, default_factory=list)

    def __post_init__(self):
        if self.wea_file != "":
            self.files_mtime.append(os.path.getmtime(self.wea_file))
        if self.epw_file != "":
            self.files_mtime.append(os.path.getmtime(self.epw_file))


@dataclass
class Model:
    """Model dataclass.

    Attributes:
        scene: SceneConfig object
        windows: A dictionary of WindowConfig
        materials: MaterialConfig object
        sensors: A dictionary of SensorConfig
        views: A dictionary of ViewConfig
    """

    materials: "MaterialConfig"
    scene: "SceneConfig" = field(default_factory=SceneConfig)
    windows: dict[str, "WindowConfig"] = field(default_factory=dict)
    sensors: dict[str, "SensorConfig"] = field(default_factory=dict)
    views: dict[str, "ViewConfig"] = field(default_factory=dict)
    surfaces: dict[str, "SurfaceConfig"] = field(default_factory=dict)

    # Make Path() out of all path strings
    def __post_init__(self):
        if isinstance(self.scene, dict):
            self.scene = SceneConfig(**self.scene)
        if isinstance(self.materials, dict):
            self.materials = MaterialConfig(**self.materials)
        for k, v in self.windows.items():
            if isinstance(v, dict):
                self.windows[k] = WindowConfig(**v)
        for k, v in self.sensors.items():
            if isinstance(v, dict):
                self.sensors[k] = SensorConfig(**v)
        for k, v in self.views.items():
            if isinstance(v, dict):
                self.views[k] = ViewConfig(**v)
        for k, v in self.surfaces.items():
            if isinstance(v, dict):
                self.surfaces[k] = SurfaceConfig(**v)

        self.scene_cfg = True
        self.windows_cfg = True
        self.sensors_cfg = True
        self.views_cfg = True
        self.surfaces_cfg = True

        if self.scene == SceneConfig() or self.scene == {}:
            self.scene_cfg = False
        if self.windows == {}:
            self.windows_cfg = False
        if self.sensors == {}:
            self.sensors_cfg = False
        if self.views == {}:
            self.views_cfg = False
        if self.surfaces == {}:
            self.surfaces_cfg = False

        # add view to sensors if not already there
        for k, v in self.views.items():
            if k in self.sensors:
                if [tuple(self.sensors[k].data[0])] == [
                    self.views[k].view.vp + self.views[k].view.vdir
                ]:
                    continue
                else:
                    raise ValueError(f"Sensor {k} data does not match view {k} data")
            else:
                self.sensors[k] = SensorConfig(
                    data=[self.views[k].view.vp + self.views[k].view.vdir]
                )

        for k, v in self.windows.items():
            if v.matrix_name != "":
                if v.matrix_name not in self.materials.matrices:
                    raise ValueError(
                        f"{k} matrix name {v.matrix_name} not found in materials"
                    )


@dataclass
class WorkflowConfig:
    """Workflow configuration dataclass.

    Workflow configuration is initialized with the Settings
    and Model dataclasses. A hash string is generated from
    the config content.

    Attributes:
        settings: A Settings object.
        model: A Model object.
        hash_str: A hash string of the config content.
    """

    settings: "Settings"
    model: "Model"
    hash_str: str = field(init=False)

    def __post_init__(self):
        if (
            not self.model.sensors_cfg
            and not self.model.views_cfg
            and not self.model.surfaces_cfg
        ):
            raise ValueError(
                f"Sensors, views, or surfaces must be specified for {self.settings.method} method"
            )
        if (
            self.settings.method == "3phase" or self.settings.method == "5phase"
        ) and not self.model.windows_cfg:
            raise ValueError(
                f"Windows must be specified in Model for the {self.settings.method} method"
            )
        if isinstance(self.settings, dict):
            self.settings = Settings(**self.settings)
        if isinstance(self.model, dict):
            self.model = Model(**self.model)
        self.hash_str = hashlib.md5(str(self.__dict__).encode()).hexdigest()[:16]

    @staticmethod
    def from_dict(obj: dict) -> "WorkflowConfig":
        """Generate a WorkflowConfig object from a dictionary.
        Args:
            obj: A dictionary of workflow configuration.
        Returns:
            A WorkflowConfig object.
        """
        settings = Settings(**obj["settings"])
        model = Model(**obj["model"])
        return WorkflowConfig(settings, model)


class PhaseMethod:
    """Base class for phase methods.

    This class is not meant to be used by itself.
    Use one of the subclasses instead.
    The base class instantiate a set of common attributes,
    along with a host of methods that are shared by all phase methods.

    Attributes:
        config: A WorkflowConfig object.
        view_senders: A dictionary of ViewSender objects.
        sensor_senders: A dictionary of SensorSender objects.
        sky_receiver: A SkyReceiver object.
        wea_header: Weather file header string
        wea_metadata: Weather file metadata object
        wea_str: Weather data string
        tmpdir: A temporary directory for storing intermediate files.
        octdir: A directory for storing octree files.
        mtxdir: A directory for storing matrix files.
        mfile: A matrix file path.
    """

    def __init__(self, config: WorkflowConfig):
        """
        Initialize a phase method.

        Args:
            config: A WorkflowConfig object.
        """
        self.config = config

        # Setup the view and sensor senders
        self.view_senders = {}
        self.sensor_senders = {}
        self.surface_senders = {}
        for name, sensors in self.config.model.sensors.items():
            self.sensor_senders[name] = SensorSender(sensors.data)
        for name, view in self.config.model.views.items():
            self.view_senders[name] = ViewSender(
                view.view, xres=view.xres, yres=view.yres
            )
        for name, surface in self.config.model.surfaces.items():
            self.surface_senders[name] = SurfaceSender(
                surfaces=surface.primitives,
                basis=surface.basis,
            )

        # Setup the sky receiver object
        self.sky_receiver = SkyReceiver(self.config.settings.sky_basis)

        # Figure out the weather related stuff
        if self.config.settings.epw_file != "":
            with open(self.config.settings.epw_file) as f:
                self.wea_metadata, self.wea_data = parse_epw(f.read())
            self.wea_header = self.wea_metadata.wea_header()
            self.wea_str = self.wea_header + "\n".join(str(d) for d in self.wea_data)
        elif self.config.settings.wea_file != "":
            with open(self.config.settings.wea_file) as f:
                self.wea_metadata, self.wea_data = parse_wea(f.read())
            self.wea_header = self.wea_metadata.wea_header()
            self.wea_str = self.wea_header + "\n".join(str(d) for d in self.wea_data)
        else:
            if (
                self.config.settings.latitude is None
                or self.config.settings.longitude is None
            ):
                raise ValueError(
                    "Latitude and longitude must be specified if no weather file is given"
                )
            self.wea_header = self.get_wea_header()
            self.wea_metadata = WeaMetaData(
                "city",
                "country",
                self.config.settings.latitude,
                self.config.settings.longitude,
                self.config.settings.time_zone,
                self.config.settings.site_elevation,
            )
            self.wea_data = None

        # Setup Temp and Octrees directory
        self.tmpdir = Path("Temp")
        self.tmpdir.mkdir(exist_ok=True)
        self.octdir = Path("Octrees")
        self.octdir.mkdir(exist_ok=True)
        self.mtxdir = Path(self.config.settings.matrix_dir)
        self.mtxdir.mkdir(exist_ok=True)
        self.mfile = (self.mtxdir / self.config.hash_str).with_suffix(".npz")

        # Generate a base octree
        self.octree = self.octdir / f"{random_string(5)}.oct"

    def __enter__(self):
        """
        Context manager enter method. This method is called when
        the class is used as a context manager. Anything happens
        after the with statement is run.
        """
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """
        Context manager exit method. This method is called when
        the class is used as a context manager. Cleans up the
        Temp and Octrees directory.
        """
        rmtree("Octrees")
        rmtree("Matrices")
        rmtree("Temp")

    def get_wea_header(self, unit: int = 1):
        meta_data = WeaMetaData(
            city=str(self.config.settings.latitude),
            country=str(self.config.settings.longitude),
            latitude=self.config.settings.latitude,
            longitude=self.config.settings.longitude,
            timezone=self.config.settings.time_zone,
            elevation=self.config.settings.site_elevation,
            unit=unit,
        )
        return meta_data.wea_header()

    def generate_matrices(self):
        raise NotImplementedError

    def calculate_view(self, view, time, dni, dhi):
        raise NotImplementedError

    def calculate_sensor(self, sensor, time, dni, dhi):
        raise NotImplementedError

    def get_sky_matrix(
        self,
        time: datetime | list[datetime],
        dni: float | list[float],
        dhi: float | list[float],
        solar_spectrum: bool = False,
        orient: float = 0.0,
    ) -> np.ndarray:
        """Generates a sky matrix based on the time, Direct Normal Irradiance (DNI), and
        Diffuse Horizontal Irradiance (DHI).

        Args:
            time: The specific time for the matrix.
            dni: The Direct Normal Irradiance value.
            dhi: The Diffuse Horizontal Irradiance value.
            solar_spectrum: Whether to use the solar spectrum or not.
            orient: The orientation of the matrix.

        Returns:
            numpy.ndarray: The generated sky matrix, with dimensions based on the
                BASIS_DIMENSION setting for the current sky_basis configuration.
        """
        _wea = self.wea_header
        _ncols = 1
        if (
            isinstance(time, datetime)
            and isinstance(dni, (float, int))
            and isinstance(dhi, (float, int))
        ):
            _wea += str(WeaData(time, dni, dhi))
        elif isinstance(time, list) and isinstance(dni, list) and isinstance(dhi, list):
            rows = [str(WeaData(t, n, d)) for t, n, d in zip(time, dni, dhi)]
            _wea += "\n".join(rows)
            _ncols = len(time)
        else:
            raise ValueError(
                "Time, DNI, and DHI must be either single values or lists of values"
            )
        smx = pr.gendaymtx(
            _wea.encode(),
            outform="d",
            mfactor=int(self.config.settings.sky_basis[-1]),
            header=False,
            solar_radiance=solar_spectrum,
            rotate=orient,
        )
        return load_binary_matrix(
            smx,
            nrows=BASIS_DIMENSION[self.config.settings.sky_basis] + 1,
            ncols=_ncols,
            ncomp=3,
            dtype="d",
        )

    def get_melanopic_sky_matrix(
        self,
        time: datetime | list[datetime],
        dni: float | list[float],
        dhi: float | list[float],
        aod: float | list[float] = 0.112,
        sky_cover: float | list[float] = 0.0,
    ) -> np.ndarray:
        """Generates a sky matrix based on the time, Direct Normal Irradiance (DNI), and
        Diffuse Horizontal Irradiance (DHI).

        Args:
            time: The specific time for the matrix.
            dni: The Direct Normal Irradiance value.
            dhi: The Diffuse Horizontal Irradiance value.

        Returns:
            numpy.ndarray: The generated sky matrix, with dimensions based on the
                BASIS_DIMENSION setting for the current sky_basis configuration.
        """
        _wea = self.get_wea_header(unit=3)
        _ncols = 1
        if (
            isinstance(time, datetime)
            and isinstance(dni, (float, int))
            and isinstance(dhi, (float, int))
            and isinstance(aod, (float, int))
            and isinstance(sky_cover, (float, int))
        ):
            _wea += str(WeaData(time, dni, dhi, aod, sky_cover))
        elif (
            isinstance(time, list)
            and isinstance(dni, list)
            and isinstance(dhi, list)
            and isinstance(aod, list)
            and isinstance(sky_cover, list)
        ):
            rows = [
                str(WeaData(t, n, d, a, s))
                for t, n, d, a, s in zip(time, dni, dhi, aod, sky_cover)
            ]
            _wea += "\n".join(rows)
            _ncols = len(time)
        smx = pr.gensdaymtx(
            _wea.encode(),
            outform="f",
            mfactor=int(self.config.settings.sky_basis[-1]),
            header=True,
            nthreads=self.config.settings.num_processors,
        )
        smx = pr.getinfo(
            pr.Rcomb(transform="m", header=False, outform="f").add_input(smx)(),
            strip_header=True,
        )
        return load_binary_matrix(
            smx,
            nrows=BASIS_DIMENSION[self.config.settings.sky_basis] + 1,
            ncols=_ncols,
            ncomp=1,
            dtype="f",
        )

    def get_sky_matrix_from_wea(self, mfactor: int, sun_only=False, onesun=False):
        if self.wea_str is None:
            raise ValueError("No weather string available")
        _sun_str = pr.gendaymtx(
            self.wea_str.encode(),
            sun_file="-",
            dryrun=True,
            daylight_hours_only=True,
        )
        prims = pr.parse_primitive(_sun_str.decode())
        _datetimes = [
            datetime(2023, 1, 1) + timedelta(int(p.identifier.lstrip("solar")))
            for p in prims
            if p.ptype == "light"
        ]
        _matrix = pr.gendaymtx(
            self.wea_str.encode(),
            sun_only=sun_only,
            onesun=onesun,
            outform="d",
            daylight_hours_only=True,
            mfactor=mfactor,
        )
        _nrows, _ncols, _ncomp, _dtype = parse_rad_header(pr.getinfo(_matrix).decode())
        return load_binary_matrix(
            _matrix,
            nrows=_nrows,
            ncols=_ncols,
            ncomp=_ncomp,
            dtype=_dtype,
            header=True,
        )

    def save_matrices(self):
        raise NotImplementedError


class TwoPhaseMethod(PhaseMethod):
    """Implements two phase method."""

    def __init__(self, config: WorkflowConfig):
        """Initializes the two phase method

        Args:
            config: A WorkflowConfig object
        """
        super().__init__(config)
        oct_stdin = config.model.materials.bytes + config.model.scene.bytes
        for window in config.model.windows.values():
            oct_stdin += window.bytes
        with open(self.octree, "wb") as f:
            f.write(
                pr.oconv(
                    *config.model.materials.files,
                    *config.model.scene.files,
                    stdin=oct_stdin,
                )
            )
        self.view_sky_matrices = {}
        self.sensor_sky_matrices = {}
        for vs in self.view_senders:
            self.view_sky_matrices[vs] = Matrix(
                self.view_senders[vs], [self.sky_receiver], self.octree
            )
        for ss in self.sensor_senders:
            self.sensor_sky_matrices[ss] = Matrix(
                self.sensor_senders[ss], [self.sky_receiver], self.octree
            )

    def generate_matrices(self) -> None:
        """Generate matrices for all view and sensor points."""
        # First check if matrices files already exist
        if self.mfile.exists() and (not self.config.settings.overwrite):
            self.load_matrices()
            return
        # Then check if overwrite is set to True
        for _, mtx in self.view_sky_matrices.items():
            mtx.generate(
                self.config.settings.view_sky_matrix,
                nproc=self.config.settings.num_processors,
            )
        for _, mtx in self.sensor_sky_matrices.items():
            mtx.generate(
                self.config.settings.sensor_sky_matrix,
                nproc=self.config.settings.num_processors,
            )
        if self.config.settings.save_matrices:
            self.save_matrices()

    def load_matrices(self):
        """
        Load matrices from a .npz file
        """
        logger.info(f"Loading matrices from {self.mfile}")
        if not self.mfile.exists():
            raise FileNotFoundError("Matrices file not found")
        mdata = np.load(self.mfile)
        for view, mtx in self.view_sky_matrices.items():
            mtx.array = mdata[f"{view}_sky_matrix"]
        for sensor, mtx in self.sensor_sky_matrices.items():
            mtx.array = mdata[f"{sensor}_sky_matrix"]

    def calculate_view(
        self, view: str, time: datetime, dni: float, dhi: float
    ) -> np.ndarray:
        """Calculate (render) a view.
        Args:
            view: A view name, must bed loaded during configuration.
            time: A datetime object
            dni: Direct normal irradiance
            dhi: Diffuse horizontal irradiance
        Returns:
            A image as a numpy array
        """
        sky_matrix = self.get_sky_matrix(time, dni, dhi)
        return matrix_multiply_rgb(self.view_sky_matrices[view].array, sky_matrix)

    def calculate_sensor(
        self, sensor: str, time: datetime, dni: float, dhi: float
    ) -> np.ndarray:
        """Calculate a sensor view.

        Args:
            sensor: A sensor name, must be loaded during configuration.
            time: A datetime object
            dni: Direct normal irradiance
            dhi: Diffuse horizontal irradiance
        Returns:
            Sensor illuminance value
        """
        sky_matrix = self.get_sky_matrix(time, dni, dhi)
        return matrix_multiply_rgb(
            self.sensor_sky_matrices[sensor].array,
            sky_matrix,
            weights=[47.4, 119.9, 11.6],
        )

    def calculate_view_from_wea(self, view: str) -> np.ndarray:
        """Render a series of images for a view. Rendering using
        the weather file loaded during configuration.

        Args:
            view: View name, must be loaded during configuration.
        Returns:
            A numpy array containing a series of images
        """
        if self.wea_data is None:
            raise ValueError("No wea data available")
        sky_matrix = self.get_sky_matrix_from_wea(
            int(self.config.settings.sky_basis[-1])
        )
        # arbitrary chunksize
        chunksize = 300
        shape = (
            self.view_sky_matrices[view].nrows,
            sky_matrix.shape[1],
            3,
        )
        final = np.memmap(
            f"{view}_2ph.dat",
            shape=shape,
            dtype=np.float64,
            mode="w+",
            order="F",
        )
        for idx in range(0, sky_matrix.shape[1], chunksize):
            end = min(idx + chunksize, sky_matrix.shape[1])
            _chunksize = end - idx
            res = matrix_multiply_rgb(
                self.view_sky_matrices[view].array,
                sky_matrix[:, idx:end, :],
            )
            final[:, idx:end, 0] = res[:, :, 0]
            final[:, idx:end, 1] = res[:, :, 1]
            final[:, idx:end, 2] = res[:, :, 2]
            final.flush()
        return final

    def calculate_sensor_from_wea(self, sensor: str) -> np.ndarray:
        """Calculate a sensor for tue duration of the weather file
        that's loaded during configuration.

        Args:
            sensor: sensor name, must be loaded during configuration.
        Returns:
            A numpy array containing a series of illuminance values
        """
        if self.wea_data is None:
            raise ValueError("No wea data available")
        return matrix_multiply_rgb(
            self.sensor_sky_matrices[sensor].array,
            self.get_sky_matrix_from_wea(int(self.config.settings.sky_basis[-1])),
            weights=[47.4, 119.9, 11.6],
        )

    def save_matrices(self):
        """Save matrices to a .npz file in the Matrices directory.
        File name is the hash string of the configuration.
        """
        matrices = {}
        for view, mtx in self.view_sky_matrices.items():
            matrices[f"{view}_sky_matrix"] = mtx.array
        for sensor, mtx in self.sensor_sky_matrices.items():
            matrices[f"{sensor}_sky_matrix"] = mtx.array
        np.savez(self.mtxdir / self.config.hash_str, **matrices)


class ThreePhaseMethod(PhaseMethod):
    """Three phase method implementation.

    Attributes:
        config: A WorkflowConfig object
        octree: A path to the octree file
        window_senders: A dictionary of window sender matrices
        window_receivers: A dictionary of window receiver matrices
        window_bsdfs: A dictionary of window BSDF matrices
        daylight_matrices: A dictionary of daylight matrices
        view_window_matrices: A dictionary of view window matrices
        sensor_window_matrices: A dictionary of sensor window matrices
    """

    def __init__(self, config):
        super().__init__(config)
        with open(self.octree, "wb") as f:
            f.write(
                pr.oconv(
                    *config.model.materials.files,
                    *config.model.scene.files,
                    stdin=config.model.materials.bytes + config.model.scene.bytes,
                )
            )
        self.window_senders: dict[str, SurfaceSender] = {}
        self.window_receivers = {}
        self.window_bsdfs = {}
        self.daylight_matrices = {}
        for _name, window in self.config.model.windows.items():
            _prims = pr.parse_primitive(window.bytes.decode())
            if window.matrix_name != "":
                self.window_bsdfs[_name] = self.config.model.materials.matrices[
                    window.matrix_name
                ].matrix_data
                window_basis = [
                    k
                    for k, v in BASIS_DIMENSION.items()
                    if v == self.window_bsdfs[_name].shape[0]
                ][0]
            else:
                # raise ValueError("No matrix data or file available", _name)
                logger.info(f"No matrix data or file available: {_name}")
            if _name in self.window_bsdfs:
                window_basis = [
                    k
                    for k, v in BASIS_DIMENSION.items()
                    if v == self.window_bsdfs[_name].shape[0]
                ][0]
            else:
                window_basis = self.config.settings.window_basis
            self.window_receivers[_name] = SurfaceReceiver(
                _prims,
                window_basis,
            )
            self.window_senders[_name] = SurfaceSender(_prims, window_basis)
            self.daylight_matrices[_name] = Matrix(
                self.window_senders[_name],
                [self.sky_receiver],
                self.octree,
            )
        self.view_window_matrices = {}
        self.sensor_window_matrices = {}
        self.surface_window_matrices = {}
        for _v, sender in self.view_senders.items():
            self.view_window_matrices[_v] = Matrix(
                sender, list(self.window_receivers.values()), self.octree
            )
        for _s, sender in self.sensor_senders.items():
            self.sensor_window_matrices[_s] = Matrix(
                sender, list(self.window_receivers.values()), self.octree
            )
        for _s, sender in self.surface_senders.items():
            self.surface_window_matrices[_s] = Matrix(
                sender, list(self.window_receivers.values()), self.octree
            )

    def generate_matrices(self, view_matrices: bool = True):
        """Generate all required matrices

        Args:
            view_matrices: Toggle to generate view matrices. Toggle it off can be
                useful for not needing the view matrices but still need the view data
                for things like edgps calculation.
        """
        if self.mfile.exists() and (not self.config.settings.overwrite):
            self.load_matrices()
            return
        if view_matrices:
            for _, mtx in self.view_window_matrices.items():
                mtx.generate(
                    self.config.settings.view_window_matrix,
                    nproc=self.config.settings.num_processors,
                )
        for _, mtx in self.sensor_window_matrices.items():
            mtx.generate(
                self.config.settings.sensor_window_matrix,
                nproc=self.config.settings.num_processors,
            )
        for _, mtx in self.surface_window_matrices.items():
            mtx.generate(
                self.config.settings.surface_window_matrix,
                nproc=self.config.settings.num_processors,
            )
        for _, mtx in self.daylight_matrices.items():
            mtx.generate(
                self.config.settings.daylight_matrix,
                nproc=self.config.settings.num_processors,
            )
        if self.config.settings.save_matrices:
            self.save_matrices()

    def load_matrices(self):
        """Load matrices from a .npz file in the Matrices directory."""
        logger.info(f"Loading matrices from {self.mfile}")
        mdata = np.load(self.mfile)
        for view, mtx in self.view_window_matrices.items():
            if (key := f"{view}_window_matrix") in mdata:
                mtx.array = mdata[key]
        for sensor, mtx in self.sensor_window_matrices.items():
            mtx.array = mdata[f"{sensor}_window_matrix"]
        for surface, mtx in self.surface_window_matrices.items():
            mtx.array = mdata[f"{surface}_window_matrix"]
        for name, mtx in self.daylight_matrices.items():
            mtx.array = mdata[f"{name}_daylight_matrix"]

    def calculate_view(
        self,
        view: str,
        bsdf: np.ndarray,
        time: datetime,
        dni: float,
        dhi: float,
    ) -> np.ndarray:
        """Calculate (render) a view.

        Args:
            view: The view name
            bsdf: The BSDF matrix
            time: The datetime object
            dni: The direct normal irradiance
            dhi: The diffuse horizontal irradiance
        Returns:
            A image as numpy array
        """
        sky_matrix = self.get_sky_matrix(time, dni, dhi)
        res = []
        if isinstance(bsdf, list):
            if len(bsdf) != len(self.config.model.windows):
                raise ValueError("Number of BSDF should match number of windows.")
        for idx, _name in enumerate(self.config.model.windows):
            _bsdf = bsdf[idx] if isinstance(bsdf, list) else bsdf
            res.append(
                matrix_multiply_rgb(
                    self.view_window_matrices[view].array[idx],
                    _bsdf,
                    self.daylight_matrices[_name].array,
                    sky_matrix,
                )
            )
        return np.sum(res, axis=0)

    def calculate_sensor(
        self,
        sensor: str,
        bsdf: dict[str, str],
        time: datetime,
        dni: float,
        dhi: float,
    ) -> np.ndarray:
        """Calculate illuminance for a sensor.

        Args:
            sensor: The sensor name
            bsdf: A dictionary of window name as key and bsdf matrix or matrix name as value
            time: The datetime object
            dni: The direct normal irradiance
            dhi: The diffuse horizontal irradiance
        Returns:
            A float value of illuminance
        """
        sky_matrix = self.get_sky_matrix(time, dni, dhi)
        res = []
        if isinstance(bsdf, list):
            if len(bsdf) != len(self.config.model.windows):
                raise ValueError("Number of BSDF should match number of windows.")
        for idx, _name in enumerate(self.config.model.windows):
            _bsdf = self.config.model.materials.matrices[bsdf[_name]].matrix_data
            res.append(
                matrix_multiply_rgb(
                    self.sensor_window_matrices[sensor].array[idx],
                    _bsdf,
                    self.daylight_matrices[_name].array,
                    sky_matrix,
                    weights=[47.4, 119.9, 11.6],
                )
            )
        return np.sum(res, axis=0)

    def calculate_view_from_wea(self, view: str) -> np.ndarray:
        """Calculate(render) view from wea data.

        Args:
            view: The view name
        Returns:
            A series of HDR images as a numpy array
        """
        if self.wea_data is None:
            raise ValueError("No wea data available")
        sky_matrix = self.get_sky_matrix_from_wea(
            int(self.config.settings.sky_basis[-1])
        )
        # arbitrary chunksize
        chunksize = 300
        shape = (
            self.view_senders[view].xres * self.view_senders[view].yres,
            sky_matrix.shape[1],
            3,
        )
        final = np.memmap(
            f"{view}_3ph.dat",
            shape=shape,
            dtype=np.float64,
            mode="w+",
            order="F",
        )
        for idx in range(0, sky_matrix.shape[1], chunksize):
            end = min(idx + chunksize, sky_matrix.shape[1])
            _chunksize = end - idx
            res = np.zeros(
                (
                    self.view_senders[view].xres * self.view_senders[view].yres,
                    _chunksize,
                    3,
                )
            )
            for widx, _name in enumerate(self.config.model.windows):
                res += matrix_multiply_rgb(
                    self.view_window_matrices[view].array[widx],
                    self.window_bsdfs[_name],
                    self.daylight_matrices[_name].array,
                    sky_matrix[:, idx:end, :],
                )
            final[:, idx:end, 0] = res[:, :, 0]
            final[:, idx:end, 1] = res[:, :, 1]
            final[:, idx:end, 2] = res[:, :, 2]
            final.flush()
        return final

    def calculate_sensor_from_wea(self, sensor: str) -> np.ndarray:
        """Calculates the sensor values from wea data.

        Args:
            sensor: The specific sensor for which the calculation is to be
                performed.

        Returns:
            numpy.ndarray: A matrix containing the calculated sensor values based on
                the Weather Attribute data, sensor configuration, and various matrices
                related to windows, daylight, and sky.

        Raises:
            ValueError: If no wea data is available.

        Examples:
            sensor_values = sensor_config.calculate_sensor_from_wea("sensor_name")
        """
        if self.wea_data is None:
            raise ValueError("No wea data available")
        sky_matrix = self.get_sky_matrix_from_wea(
            int(self.config.settings.sky_basis[-1])
        )
        res = np.zeros((self.sensor_senders[sensor].yres, sky_matrix.shape[1]))
        for idx, _name in enumerate(self.config.model.windows):
            res += matrix_multiply_rgb(
                self.sensor_window_matrices[sensor].array[idx],
                self.window_bsdfs[_name],
                self.daylight_matrices[_name].array,
                sky_matrix,
                weights=[47.4, 119.9, 11.6],
            )
        return res

    def calculate_surface(
        self,
        surface: str,
        bsdf: dict[str, str],
        time: datetime,
        dni: float,
        dhi: float,
        solar_spectrum: bool = False,
        sky_matrix: None | np.ndarray = None,
    ) -> np.ndarray:
        weights = [47.4, 119.9, 11.6]
        if solar_spectrum:
            weights = [1.0, 1.0, 1.0]
        if sky_matrix is None:
            sky_matrix = self.get_sky_matrix(
                time, dni, dhi, solar_spectrum=solar_spectrum
            )
        res = np.zeros((self.surface_senders[surface].yres, sky_matrix.shape[1]))
        for idx, _name in enumerate(self.config.model.windows):
            _bsdf = self.config.model.materials.matrices[bsdf[_name]].matrix_data
            res += matrix_multiply_rgb(
                self.surface_window_matrices[surface].array[idx],
                _bsdf,
                self.daylight_matrices[_name].array,
                sky_matrix,
                weights=weights,
            )
        return res

    def calculate_mev(
        self,
        sensor: str,
        bsdf: dict[str, str],
        time: datetime,
        dni: float,
        dhi: float,
        sky_cover: float,
    ) -> np.ndarray:
        """Calculate menalonpic vertical illuminance.

        Args:
            sensor: sensor name, must be in config.model.sensors
            bsdf: a dictionary of window name as key and bsdf matrix or matrix name as value
            time: datetime object
            dni: direct normal irradiance
            dhi: diffuse horizontal irradiance
            sky_cover: sky cover fraction
        Returns:
            Menalonpic vertical illuminance
        """
        sky_matrix = self.get_melanopic_sky_matrix(time, dni, dhi, sky_cover=sky_cover)
        res = []
        if isinstance(bsdf, list):
            if len(bsdf) != len(self.config.model.windows):
                raise ValueError("Number of BSDF should match number of windows.")
        for idx, _name in enumerate(self.config.model.windows):
            matrix_name = bsdf[_name]
            _bsdf = self.config.model.materials.matrices_mlnp[matrix_name].matrix_data
            _vmx = self.sensor_window_matrices[sensor].array[idx][:, :, 0]
            _dmx = self.daylight_matrices[_name].array[:, :, 0]
            _smx = sky_matrix[:, :, 0]
            res.append(np.linalg.multi_dot([_vmx, _bsdf, _dmx, _smx]))
        return np.sum(res, axis=0)

    def calculate_edgps(
        self,
        view: str,
        bsdf: dict[str, str],
        time: datetime,
        dni: float,
        dhi: float,
        ambient_bounce: int = 0,
        save_hdr: None | str | Path = None,
    ) -> tuple[float, float]:
        """Calculate enhanced simplified daylight glare probability (EDGPs) for a view.

        Args:
            view: view name, must be in config.model.views
            bsdf: a dictionary of window name as key and bsdf matrix or matrix name as value
            time: datetime object
            dni: direct normal irradiance
            dhi: diffuse horizontal irradiance
            ambient_bounce: ambient bounce, default to 1. Could be set to zero for
                macroscopic non-scattering systems.
        Returns:
            EDGPs
        """
        # generate octree with bsdf
        stdins = []
        stdins.append(
            gen_perez_sky(
                time,
                self.wea_metadata.latitude,
                self.wea_metadata.longitude,
                self.wea_metadata.timezone,
                dirnorm=dni,
                diffhor=dhi,
            )
        )
        for wname, sname in bsdf.items():
            if (_pgs := self.config.model.windows[wname].proxy_geometry) != {}:
                stdins.append(_pgs[sname])

        octree = f"{random_string(5)}.oct"
        with open(octree, "wb") as f:
            f.write(pr.oconv(stdin=b"".join(stdins), octree=self.octree))

        # render image with -ab 1
        params = ["-ab", str(ambient_bounce)]
        hdr = pr.rpict(
            pr.get_view_args(self.view_senders[view].view),
            octree,
            xres=800,
            yres=800,
            params=params,
        )
        ev = self.calculate_sensor(
            view,
            bsdf,
            time,
            dni,
            dhi,
        )
        if save_hdr is not None:
            with open(save_hdr, "wb") as f:
                f.write(hdr)
        res = pr.evalglare(hdr, fast=1, correction_mode="l-", ev=ev.item())
        edgps = float(res)
        os.remove(octree)
        return edgps, ev.item()

    def save_matrices(self):
        """Saves the view window matrices, sensor window matrices, and daylight matrices
        to a NumPy `.npz` file.

        The matrices are saved with keys formed by concatenating the corresponding
        view, sensor, or window name with '_window_matrix' or '_daylight_matrix'.
        """
        matrices = {}
        for view, mtx in self.view_window_matrices.items():
            matrices[f"{view}_window_matrix"] = mtx.array
        for sensor, mtx in self.sensor_window_matrices.items():
            matrices[f"{sensor}_window_matrix"] = mtx.array
        for surface, mtx in self.surface_window_matrices.items():
            matrices[f"{surface}_window_matrix"] = mtx.array
        for window, mtx in self.daylight_matrices.items():
            matrices[f"{window}_daylight_matrix"] = mtx.array
        np.savez(self.mfile, **matrices)


class FivePhaseMethod(PhaseMethod):
    """
    A class representing the Five-Phase Method, which is an extension of the
    three-phase method, allowing for more complex simulations. It includes
    various matrices, octrees, and other attributes used in the simulation.

    Attributes:
        blacked_out_octree: Path to the octree with blacked-out surfaces.
        vmap_oct: Path to the vmap octree.
        cdmap_oct: Path to the cdmap octree.
        window_senders: dictionary of window sender objects.
        window_receivers: dictionary of window receiver objects.
        window_bsdfs: dictionary of window BSDFs.
        view_window_matrices: dictionary of view window matrices.
        sensor_window_matrices: dictionary of sensor window matrices.
        daylight_matrices: dictionary of daylight matrices.
        direct_sun_matrix: Direct sun matrix.
    """

    def __init__(self, config: WorkflowConfig):
        """
        Initializes the FivePhaseMethod object by setting up octrees, matrices,
        and other necessary attributes.

        Reads materials and scene files, constructs necessary octrees, and
        prepares window objects, sun receivers, and various mapping matrices
        based on the provided configuration.

        Args:
            config: WorkflowConfig object containing all necessary information for
                initializing the five-phase method.
        """
        super().__init__(config)
        with open(self.octree, "wb") as f:
            f.write(
                pr.oconv(
                    *config.model.materials.files,
                    *config.model.scene.files,
                    stdin=(config.model.materials.bytes + config.model.scene.bytes),
                )
            )
        self.blacked_out_octree: Path = self.octdir / f"{random_string(5)}.oct"
        self.vmap_oct: Path = self.octdir / f"vmap_{random_string(5)}.oct"
        self.cdmap_oct: Path = self.octdir / f"cdmap_{random_string(5)}.oct"
        self.window_senders: dict[str, SurfaceSender] = {}
        self.window_receivers: dict[str, SurfaceReceiver] = {}
        self.window_bsdfs: dict[str, np.ndarray] = {}
        self.view_window_matrices: dict[str, Matrix] = {}
        self.sensor_window_matrices: dict[str, Matrix] = {}
        self.daylight_matrices: dict[str, Matrix] = {}
        self.view_window_direct_matrices: dict[str, Matrix] = {}
        self.sensor_window_direct_matrices: dict[str, Matrix] = {}
        self.daylight_direct_matrices: dict[str, Matrix] = {}
        self.sensor_sun_direct_matrices: dict[str, SunMatrix] = {}
        self.view_sun_direct_matrices: dict[str, SunMatrix] = {}
        self.view_sun_direct_illuminance_matrices: dict[str, SunMatrix] = {}
        self.vmap: dict[str, np.ndarray] = {}
        self.cdmap: dict[str, np.ndarray] = {}
        self.direct_sun_matrix: np.ndarray = self.get_sky_matrix_from_wea(
            mfactor=int(self.config.settings.sun_basis[-1]),
            onesun=True,
            sun_only=True,
        )
        self._prepare_window_objects()
        self._prepare_sun_receivers()
        self.direct_sun_matrix = to_sparse_matrix3(self.direct_sun_matrix)
        self._gen_blacked_out_octree()
        self._prepare_mapping_octrees()
        self._prepare_view_sender_objects()
        self._prepare_sensor_sender_objects()

    def _gen_blacked_out_octree(self):
        black_scene = b"\n".join(
            pr.Xform(s, modifier="black")() for s in self.config.model.scene.files
        )
        if self.config.model.scene.bytes != b"":
            black_scene += pr.Xform(self.config.model.scene.bytes, modifier="black")()
        black = pr.Primitive("void", "plastic", "black", [], [0, 0, 0, 0, 0])
        glow = pr.Primitive("void", "glow", "glowing", [], [1, 1, 1, 0])
        with open(self.blacked_out_octree, "wb") as f:
            f.write(
                pr.oconv(
                    *self.config.model.materials.files,
                    # *self.config.model.windows,
                    stdin=self.config.model.materials.bytes
                    + glow.bytes
                    + black.bytes
                    + black_scene,
                )
            )

    def _prepare_window_objects(self):
        for _name, window in self.config.model.windows.items():
            _prims = pr.parse_primitive(window.bytes)
            self.window_receivers[_name] = SurfaceReceiver(
                _prims, self.config.settings.window_basis
            )
            self.window_senders[_name] = SurfaceSender(
                _prims, self.config.settings.window_basis
            )
            if window.matrix_name != "":
                self.window_bsdfs[_name] = self.config.model.materials.matrices[
                    window.matrix_name
                ].matrix_data
            elif window.matrix_data != []:
                self.window_bsdfs[_name] = np.array(window.matrix_data)
            else:
                raise ValueError("No matrix data or file available", _name)
            self.daylight_matrices[_name] = Matrix(
                self.window_senders[_name],
                [self.sky_receiver],
                self.octree,
            )
            self.daylight_direct_matrices[_name] = Matrix(
                self.window_senders[_name],
                [self.sky_receiver],
                self.blacked_out_octree,
            )

    def _prepare_view_sender_objects(self):
        for _v, sender in self.view_senders.items():
            self.vmap[_v] = load_binary_matrix(
                pr.rtrace(
                    sender.content,
                    params=["-ffd", "-av", ".31831", ".31831", ".31831"],
                    octree=self.vmap_oct,
                ),
                nrows=sender.xres * sender.yres,
                ncols=1,
                ncomp=3,
                dtype="d",
                header=True,
            )
            self.cdmap[_v] = load_binary_matrix(
                pr.rtrace(
                    sender.content,
                    params=["-ffd", "-av", ".31831", ".31831", ".31831"],
                    octree=self.cdmap_oct,
                ),
                nrows=sender.xres * sender.yres,
                ncols=1,
                ncomp=3,
                dtype="d",
                header=True,
            )
            self.view_window_matrices[_v] = Matrix(
                sender, list(self.window_receivers.values()), self.octree
            )
            self.view_window_direct_matrices[_v] = Matrix(
                sender,
                list(self.window_receivers.values()),
                self.blacked_out_octree,
            )
            self.view_sun_direct_matrices[_v] = SunMatrix(
                sender, self.view_sun_receiver, self.blacked_out_octree
            )
            self.view_sun_direct_illuminance_matrices[_v] = SunMatrix(
                sender, self.view_sun_receiver, self.blacked_out_octree
            )

    def _prepare_sensor_sender_objects(self):
        for _s, sender in self.sensor_senders.items():
            self.sensor_window_matrices[_s] = Matrix(
                sender, list(self.window_receivers.values()), self.octree
            )
            self.sensor_window_direct_matrices[_s] = Matrix(
                sender,
                list(self.window_receivers.values()),
                self.blacked_out_octree,
            )
            self.sensor_sun_direct_matrices[_s] = SunMatrix(
                sender, self.sensor_sun_receiver, self.blacked_out_octree
            )

    def _prepare_sun_receivers(self):
        if self.config.settings.sun_culling:
            window_normals = [
                parse_polygon(r.surfaces[0]).normal.tobytes()
                for r in self.window_receivers.values()
            ]
            unique_window_normals = [np.frombuffer(arr) for arr in set(window_normals)]
            self.sensor_sun_receiver = SunReceiver(
                self.config.settings.sun_basis,
                sun_matrix=self.direct_sun_matrix,
                full_mod=True,
            )
            self.view_sun_receiver = SunReceiver(
                self.config.settings.sun_basis,
                sun_matrix=self.direct_sun_matrix,
                window_normals=unique_window_normals,
                full_mod=False,
            )
        else:
            self.sensor_sun_receiver = SunReceiver(
                self.config.settings.sun_basis, full_mod=True
            )
            self.view_sun_receiver = SunReceiver(
                self.config.settings.sun_basis, full_mod=False
            )

    def _prepare_mapping_octrees(self):
        blacked_out_windows = []
        glowing_windows = []
        for _, sender in self.window_senders.items():
            for window in sender.surfaces:
                blacked_out_windows.append(
                    str(
                        pr.Primitive(
                            "black",
                            window.ptype,
                            window.identifier,
                            window.sargs,
                            window.fargs,
                        )
                    )
                )
                glowing_windows.append(
                    str(
                        pr.Primitive(
                            "glowing",
                            window.ptype,
                            window.identifier,
                            window.sargs,
                            window.fargs,
                        )
                    )
                )
        black = pr.Primitive("void", "plastic", "black", [], [0, 0, 0, 0, 0])
        glow = pr.Primitive("void", "glow", "glowing", [], [1, 1, 1, 0])
        blacked_out_windows = str(black) + " ".join(blacked_out_windows)
        glowing_windows = str(glow) + " ".join(glowing_windows)
        with open(self.vmap_oct, "wb") as wtr:
            wtr.write(pr.oconv(stdin=glowing_windows.encode(), octree=self.octree))
        logger.info("Generating view matrix material map octree")
        with open(self.cdmap_oct, "wb") as wtr:
            wtr.write(pr.oconv(stdin=blacked_out_windows.encode(), octree=self.octree))

    def generate_matrices(self):
        if self.mfile.exists():
            if not self.config.settings.overwrite:
                self.load_matrices()
                return
        logger.info("Generating matrices...")
        logger.info("Step 1/5: Generating view matrices...")
        for mtx in self.view_window_matrices.values():
            mtx.generate(
                self.config.settings.view_window_matrix,
                memmap=True,
                nproc=self.config.settings.num_processors,
            )
        for mtx in self.sensor_window_matrices.values():
            mtx.generate(
                self.config.settings.sensor_window_matrix,
                nproc=self.config.settings.num_processors,
            )
        logger.info("Step 2/5: Generating daylight matrices...")
        for mtx in self.daylight_matrices.values():
            mtx.generate(
                self.config.settings.daylight_matrix,
                nproc=self.config.settings.num_processors,
            )
        logger.info("Step 3/5: Generating direct view matrices...")
        for _, mtx in self.view_window_direct_matrices.items():
            mtx.generate(["-ab", "1"], sparse=True)
        for _, mtx in self.sensor_window_direct_matrices.items():
            mtx.generate(["-ab", "1"], sparse=True)
        logger.info("Step 4/5: Generating direct daylight matrices...")
        for _, mtx in self.daylight_direct_matrices.items():
            mtx.generate(["-ab", "0"], sparse=True)
        logger.info("Step 5/5: Generating direct sun matrices...")
        for _, mtx in self.sensor_sun_direct_matrices.items():
            mtx.generate(["-ab", "0"])
        for _, mtx in self.view_sun_direct_matrices.items():
            mtx.generate(["-ab", "0"])
        for _, mtx in self.view_sun_direct_illuminance_matrices.items():
            mtx.generate(["-ab", "0", "-i+"])
        logger.info("Done!")
        if self.config.settings.save_matrices:
            self.save_matrices()

    def load_matrices(self):
        """ """
        logger.info(f"Loading matrices from {self.mfile}")
        mdata = np.load(self.mfile, allow_pickle=True)
        for view, mtx in self.view_window_matrices.items():
            mtx.array = mdata[f"{view}_window_matrix"]
        for sensor, mtx in self.sensor_window_matrices.items():
            mtx.array = mdata[f"{sensor}_window_matrix"]
        for window, mtx in self.daylight_matrices.items():
            mtx.array = mdata[f"{window}_daylight_matrix"]
        for view, mtx in self.view_window_direct_matrices.items():
            mtx.array = mdata[f"{view}_window_direct_matrix"]
        for sensor, mtx in self.sensor_window_direct_matrices.items():
            mtx.array = mdata[f"{sensor}_window_direct_matrix"]
        for window, mtx in self.daylight_direct_matrices.items():
            mtx.array = mdata[f"{window}_daylight_direct_matrix"]
        for sensor, mtx in self.sensor_sun_direct_matrices.items():
            mtx.array = mdata[f"{sensor}_sun_direct_matrix"]
        for view, mtx in self.view_sun_direct_matrices.items():
            mtx.array = mdata[f"{view}_sun_direct_matrix"]
        for view, mtx in self.view_sun_direct_illuminance_matrices.items():
            mtx.array = mdata[f"{view}_sun_direct_illuminance_matrix"]

    def calculate_view_from_wea(self, view: str):
        logger.info("Step 1/2: Generating sky matrix from wea")
        sky_matrix = self.get_sky_matrix_from_wea(
            int(self.config.settings.sky_basis[-1])
        )
        direct_sky_matrix = self.get_sky_matrix_from_wea(
            int(self.config.settings.sky_basis[-1]), sun_only=True
        )
        direct_sky_matrix = to_sparse_matrix3(direct_sky_matrix)
        logger.info("Step 2/2: Multiplying matrices...")
        chunksize = 300
        shape = (
            self.view_window_matrices[view].nrows,
            sky_matrix.shape[1],
            3,
        )
        res = np.memmap(
            f"{view}_5ph.dat",
            shape=shape,
            dtype=np.float64,
            mode="w+",
            order="F",
        )
        for idx in range(0, sky_matrix.shape[1], chunksize):
            end = min(idx + chunksize, sky_matrix.shape[1])
            _res = [[], [], []]
            for widx, _name in enumerate(self.config.model.windows):
                for c in range(3):
                    tdmx = np.dot(
                        self.window_bsdfs[_name][:, :, c],
                        self.daylight_matrices[_name].array[:, :, c],
                    )
                    tdsmx = np.dot(tdmx, sky_matrix[:, idx:end, c])
                    vtdsmx = np.dot(
                        self.view_window_matrices[view].array[widx][:, :, c],
                        tdsmx,
                    )
                    tdmx = np.dot(
                        csr_matrix(self.window_bsdfs[_name][:, :, c]),
                        self.daylight_direct_matrices[_name].array[c],
                    )
                    tdsmx = np.dot(tdmx, direct_sky_matrix[c][:, idx:end])
                    vtdsmx_d = np.dot(
                        self.view_window_direct_matrices[view].array[widx][c],
                        tdsmx,
                    )
                    _res[c].append(vtdsmx - vtdsmx_d.toarray())
            for c in range(3):
                cdr = np.dot(
                    self.view_sun_direct_matrices[view].array[c],
                    self.direct_sun_matrix[c][:, idx:end],
                )
                cdf = np.dot(
                    self.view_sun_direct_illuminance_matrices[view].array[c],
                    self.direct_sun_matrix[c][:, idx:end],
                ).multiply(csr_matrix(self.cdmap[view][:, :, c]))
                res[:, idx:end, c] = (
                    np.sum(_res[c], axis=0) + cdr.toarray() + cdf.toarray()
                )
            res.flush()
        return res

    def calculate_sensor_from_wea(self, sensor):
        sky_matrix = self.get_sky_matrix_from_wea(
            int(self.config.settings.sky_basis[-1])
        )
        direct_sky_matrix = self.get_sky_matrix_from_wea(
            int(self.config.settings.sky_basis[-1]), sun_only=True
        )
        direct_sky_matrix = to_sparse_matrix3(direct_sky_matrix)
        res3 = np.zeros((self.sensor_senders[sensor].yres, sky_matrix.shape[1]))
        res3d = np.zeros((self.sensor_senders[sensor].yres, sky_matrix.shape[1]))
        for idx, _name in enumerate(self.config.model.windows):
            res3 += matrix_multiply_rgb(
                self.sensor_window_matrices[sensor].array[idx],
                self.window_bsdfs[_name],
                self.daylight_matrices[_name].array,
                sky_matrix,
                weights=[47.4, 119.9, 11.6],
            )
            res3d += sparse_matrix_multiply_rgb_vtds(
                self.sensor_window_direct_matrices[sensor].array[idx],
                self.window_bsdfs[_name],
                self.daylight_direct_matrices[_name].array,
                direct_sky_matrix,
                weights=[47.4, 119.9, 11.6],
            )
        rescd = np.zeros((self.sensor_senders[sensor].yres, sky_matrix.shape[1]))
        for c, w in enumerate([47.4, 119.9, 11.6]):
            rescd += w * np.dot(
                self.sensor_sun_direct_matrices[sensor].array[c],
                self.direct_sun_matrix[c],
            )
        return res3 - res3d + rescd

    def save_matrices(self):
        matrices = {}
        for view, mtx in self.view_window_matrices.items():
            matrices[f"{view}_window_matrix"] = mtx.array
        for sensor, mtx in self.sensor_window_matrices.items():
            matrices[f"{sensor}_window_matrix"] = mtx.array
        for window, mtx in self.daylight_matrices.items():
            matrices[f"{window}_daylight_matrix"] = mtx.array
        for view, mtx in self.view_window_direct_matrices.items():
            matrices[f"{view}_window_direct_matrix"] = mtx.array
        for sensor, mtx in self.sensor_window_direct_matrices.items():
            matrices[f"{sensor}_window_direct_matrix"] = mtx.array
        for window, mtx in self.daylight_direct_matrices.items():
            matrices[f"{window}_daylight_direct_matrix"] = mtx.array
        for sensor, mtx in self.sensor_sun_direct_matrices.items():
            matrices[f"{sensor}_sun_direct_matrix"] = mtx.array
        for view, mtx in self.view_sun_direct_matrices.items():
            matrices[f"{view}_sun_direct_matrix"] = mtx.array
        for view, mtx in self.view_sun_direct_illuminance_matrices.items():
            matrices[f"{view}_sun_direct_illuminance_matrix"] = mtx.array
        np.savez_compressed(self.mfile, **matrices)


def get_workflow(config):
    workflow = None
    if config.settings.method.lower().startswith(("2", "two")):
        workflow = TwoPhaseMethod(config)
    elif config.settings.method.lower().startswith(("3", "three")):
        workflow = ThreePhaseMethod(config)
    elif config.settings.method.lower().startswith(("5", "five")):
        workflow = FivePhaseMethod(config)
    else:
        raise NotImplementedError("Method not implemented")
    return workflow



import os
from pathlib import Path
import unittest

from frads import utils
from pyradiance import Primitive


class TestUtils(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        test_dir_path = Path(os.path.dirname(__file__))
        resource_dir = test_dir_path / "Resources"
        cls.prim_path = resource_dir / "model" / "Objects" / "floor_openroom.rad"

    def test_unpack_primitives(self):
        prims = utils.unpack_primitives(self.prim_path)
        self.assertIsInstance(prims[0], Primitive)



import json
import os
import math
import tempfile
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import Sequence

import numpy as np
import pyradiance as pr
import pywincalc as pwc

from frads.geom import Polygon, angle_between, polygon_primitive
from frads.utils import parse_primitive

AIR = pwc.PredefinedGasType.AIR
KRYPTON = pwc.PredefinedGasType.KRYPTON
XENON = pwc.PredefinedGasType.XENON
ARGON = pwc.PredefinedGasType.ARGON

M_PER_MM = 0.001
NM_PER_MM = 1e3


SHADING = "shading"
GLAZING = "glazing"
VENETIAN = "venetian"
FABRIC = "fabric"


@dataclass(slots=True)
class PaneRGB:
    """Pane color data object.

    Attributes:
        measured_data: measured data as a PaneProperty object.
        coated_rgb: Coated side RGB.
        glass_rgb: Non-coated side RGB.
        trans_rgb: Transmittance RGB.
    """

    coated_rgb: tuple[float, float, float]
    glass_rgb: tuple[float, float, float]
    trans_rgb: tuple[float, float, float]
    coated_side: None | str = None


@dataclass(slots=True)
class OpeningDefinitions:
    top_m: float = 0.0
    bottom_m: float = 0.0
    left_m: float = 0.0
    right_m: float = 0.0
    front_multiplier: float = 0.05


@dataclass(slots=True)
class OpeningMultipliers:
    top: float = 0.0
    bottom: float = 0.0
    left: float = 0.0
    right: float = 0.0
    front: float = 0.0


@dataclass(slots=True)
class LayerInput:
    input_source: Path | str | bytes
    flipped: bool = False
    slat_angle_deg: float = 0.0
    openings: OpeningDefinitions = field(default_factory=OpeningDefinitions)


@dataclass(slots=True)
class Layer:
    """Layer data object.

    Attributes:
        product_name: Name of the product.
        thickness: Thickness of the layer.
        product_type: Type of product.
        conductivity: Conductivity of the layer.
        emissivity_front: Front emissivity of the layer.
        emissivity_back: Back emissivity of the layer.
        ir_transmittance: IR transmittance of the layer.
        rgb: PaneRGB object.
    """

    product_name: str
    thickness_m: float
    product_type: str
    conductivity: float
    emissivity_front: float
    emissivity_back: float
    ir_transmittance: float
    flipped: bool = False
    spectral_data: dict = field(default_factory=dict)
    coated_side: str = ""
    shading_material: pr.ShadingMaterial = field(default_factory=pr.ShadingMaterial)
    slat_width_m: float = 0.0160
    slat_spacing_m: float = 0.0120
    slat_thickness_m: float = 0.0006
    slat_curve_m: float = 0.0
    slat_angle_deg: float = 90.0
    slat_conductivity: float = 160.00
    nslats: int = 1
    opening_multipliers: OpeningMultipliers = field(default_factory=OpeningMultipliers)
    shading_xml: str = ""


@dataclass(slots=True)
class Gas:
    """Gas data object.

    Attributes:
        gas: Gas type.
        ratio: Gas ratio.
    """

    gas: str
    ratio: float

    def __post_init__(self):
        if self.ratio < 0 or self.ratio > 1:
            raise ValueError("Gas ratio must be between 0 and 1.")
        if self.gas.lower() not in ("air", "argon", "krypton", "xenon"):
            raise ValueError("Invalid gas type.")


@dataclass(slots=True)
class Gap:
    """Gap data object.

    Attributes:
        gas: List of Gas objects.
        thickness: Thickness of the gap.
    """

    gas: list[Gas]
    thickness_m: float

    def __post_init__(self):
        if self.thickness_m <= 0:
            raise ValueError("Gap thickness must be greater than 0.")
        if sum(g.ratio for g in self.gas) != 1:
            raise ValueError("The sum of the gas ratios must be 1.")


@dataclass(slots=True)
class GlazingSystem:
    """Glazing system data object.

    Attributes:
        name: Name of the glazing system.
        thickness: Thickness of the glazing system.
        layers: List of Layer objects.
        gaps: List of Gap objects.
        visible_front_transmittance: Visible front transmittance matrix.
        visible_back_transmittance: Visible back transmittance matrix.
        visible_front_reflectance: Visible front reflectance matrix.
        visible_back_reflectance: Visible back reflectance matrix.
        solar_front_transmittance: Solar front transmittance matrix.
        solar_back_transmittance: Solar back transmittance matrix.
        solar_front_reflectance: Solar front reflectance matrix.
        solar_back_reflectance: Solar back reflectance matrix.
        solar_front_absorptance: Solar front absorptance matrix by layer.
        solar_back_absorptance: Solar back absorptance matrix by layer.
    """

    name: str
    thickness: float = 0
    layers: list[Layer] = field(default_factory=list)
    gaps: list[Gap] = field(default_factory=list)
    visible_front_transmittance: list[list[float]] = field(default_factory=list)
    visible_back_transmittance: list[list[float]] = field(default_factory=list)
    visible_front_reflectance: list[list[float]] = field(default_factory=list)
    visible_back_reflectance: list[list[float]] = field(default_factory=list)
    solar_front_transmittance: list[list[float]] = field(default_factory=list)
    solar_back_transmittance: list[list[float]] = field(default_factory=list)
    solar_front_reflectance: list[list[float]] = field(default_factory=list)
    solar_back_reflectance: list[list[float]] = field(default_factory=list)
    solar_front_absorptance: list[list[float]] = field(default_factory=list)
    solar_back_absorptance: list[list[float]] = field(default_factory=list)
    melanopic_back_transmittance: list[list[float]] = field(default_factory=list)
    melanopic_back_reflectance: list[list[float]] = field(default_factory=list)

    def _matrix_to_str(self, matrix: list[list[float]]) -> str:
        """Convert a matrix to a string."""
        return "\n".join([" ".join([str(n) for n in row]) for row in matrix])

    def to_xml(self, out):
        """Save the glazing system to a file."""
        with tempfile.TemporaryDirectory() as tmpdir:
            _tmpdir = Path(tmpdir)
            with open(_tvf := _tmpdir / "tbv", "w") as f:
                f.write(self._matrix_to_str(self.visible_front_transmittance))
            with open(_tvb := _tmpdir / "tfv", "w") as f:
                f.write(self._matrix_to_str(self.visible_back_transmittance))
            with open(_rvf := _tmpdir / "rfv", "w") as f:
                f.write(self._matrix_to_str(self.visible_front_reflectance))
            with open(_rvb := _tmpdir / "rbv", "w") as f:
                f.write(self._matrix_to_str(self.visible_back_reflectance))
            with open(_tsf := _tmpdir / "tbs", "w") as f:
                f.write(self._matrix_to_str(self.solar_front_transmittance))
            with open(_tsb := _tmpdir / "tfs", "w") as f:
                f.write(self._matrix_to_str(self.solar_back_transmittance))
            with open(_rsf := _tmpdir / "rfs", "w") as f:
                f.write(self._matrix_to_str(self.solar_front_reflectance))
            with open(_rsb := _tmpdir / "rbs", "w") as f:
                f.write(self._matrix_to_str(self.solar_back_reflectance))
            with open(out, "wb") as f:
                f.write(
                    pr.WrapBSDF(
                        enforce_window=True,
                        basis="kf",
                        n=self.name,
                        m="",
                        t=str(self.thickness),
                    )
                    .add_visible(str(_tvb), str(_tvf), str(_rvb), str(_rvf))
                    .add_solar(str(_tsb), str(_tsf), str(_rsb), str(_rsf))()
                )

    def save(self, out: str | Path):
        """Save the glazing system to a file."""
        out = Path(out)
        if out.suffix == ".xml":
            self.to_xml(out)
        else:
            with open(out.with_suffix(".json"), "w") as f:
                json.dump(asdict(self), f)

    @classmethod
    def from_json(cls, path: str | Path):
        """Load a glazing system from a JSON file."""
        print("\nTo be deprected please use load_glazing_system(fpath) instead.\n")
        with open(path, "r") as f:
            data = json.load(f)
        layers = data.pop("layers")
        layer_instances = []
        for layer in layers:
            smat = layer.pop("shading_material")
            shading_material = pr.ShadingMaterial(**smat)

            # Handle opening_multipliers nested dataclass
            opening_multipliers_data = layer.get("opening_multipliers", {})
            if isinstance(opening_multipliers_data, dict):
                opening_multipliers = OpeningMultipliers(**opening_multipliers_data)
            else:
                opening_multipliers = OpeningMultipliers()

            layer_instances.append(
                Layer(
                    product_name=layer["product_name"],
                    thickness_m=layer["thickness_m"],
                    product_type=layer["product_type"],
                    conductivity=layer["conductivity"],
                    emissivity_front=layer["emissivity_front"],
                    emissivity_back=layer["emissivity_back"],
                    ir_transmittance=layer["ir_transmittance"],
                    flipped=layer.get("flipped", False),
                    spectral_data={
                        int(k): v for k, v in layer["spectral_data"].items()
                    },
                    coated_side=layer.get("coated_side", ""),
                    shading_material=shading_material,
                    slat_width_m=layer.get("slat_width_m", 0.0160),
                    slat_spacing_m=layer.get("slat_spacing_m", 0.0120),
                    slat_thickness_m=layer.get("slat_thickness_m", 0.0006),
                    slat_curve_m=layer.get("slat_curve_m", 0.0),
                    slat_angle_deg=layer.get("slat_angle_deg", 90.0),
                    slat_conductivity=layer.get("slat_conductivity", 160.00),
                    nslats=layer.get("nslats", 1),
                    opening_multipliers=opening_multipliers,
                    shading_xml=layer.get("shading_xml", ""),
                )
            )
        gaps = data.pop("gaps")
        gap_instances = []
        for gap in gaps:
            gases = gap.pop("gas")
            gas_instances = [Gas(gas=gas["gas"], ratio=gas["ratio"]) for gas in gases]
            gap_instances.append(
                Gap(
                    gas=gas_instances,
                    thickness_m=gap["thickness_m"],
                    **{k: v for k, v in gap.items() if k not in ["gas", "thickness_m"]},
                )
            )
        return cls(layers=layer_instances, gaps=gap_instances, **data)

    def get_brtdfunc(self, name: None | str = None) -> None:
        """Get a BRTDfunc primitive for the glazing system."""
        print("Deprecated: use get get_glazing_brtdfunc() instead.")
        return


def get_layer_data(inp: pwc.ProductData) -> Layer:
    """Create a list of layers from a list of pwc.ProductData."""
    layer = Layer(
        product_name=inp.product_name,
        product_type=inp.product_type,
        thickness_m=inp.thickness,
        conductivity=inp.conductivity,
        emissivity_front=inp.emissivity_front,
        emissivity_back=inp.emissivity_back,
        ir_transmittance=inp.ir_transmittance,
    )
    if layer.thickness_m is not None:
        layer.thickness_m *= M_PER_MM
    return layer


def generate_blinds_xml(
    depth: float,
    nslats: int,
    angle: float,
    curvature: float,
    thickness: float,
    name: str,
    r_sol_diff: float,
    r_sol_spec: float,
    r_vis_diff: float,
    r_vis_spec: float,
    r_ir: float,
    nproc: int,
    nsamp: int,
) -> bytes:
    width = 1
    height = 1
    material_solar = pr.ShadingMaterial(r_sol_diff, r_sol_spec, 0)
    material_visible = pr.ShadingMaterial(r_vis_diff, r_vis_spec, 0)
    material_ir = pr.ShadingMaterial(r_ir, 0, 0)
    geom = pr.BlindsGeometry(
        depth=depth,
        width=width,
        height=height,
        nslats=nslats,
        angle=angle,
        rcurv=curvature,
    )
    sol_blinds = pr.generate_blinds_for_bsdf(material_solar, geom)
    vis_blinds = pr.generate_blinds_for_bsdf(material_visible, geom)
    ir_blinds = pr.generate_blinds_for_bsdf(material_ir, geom)
    spacing = height / nslats
    dim = pr.genbsdf.SamplingBox(0.5, 0.51, 0.5, 0.5 + spacing, -thickness, 0)
    sol_results = pr.generate_bsdf(sol_blinds, dim=dim, nproc=nproc, nsamp=nsamp)
    vis_results = pr.generate_bsdf(vis_blinds, dim=dim, nproc=nproc, nsamp=nsamp)
    ir_results = pr.generate_bsdf(ir_blinds, dim=dim, basis="u")
    return pr.generate_xml(
        sol_results,
        vis_results,
        ir_results,
        n=name,
        m="unnamed",
        t=thickness,
        tir=float(ir_results.front.transmittance.decode()),
    )


def load_glazing_system(path: str | Path) -> GlazingSystem:
    """Load a glazing system from a JSON file."""
    with open(path, "r") as f:
        data = json.load(f)
    layers = data.pop("layers")
    layer_instances = []
    for layer in layers:
        smat = layer.pop("shading_material")
        shading_material = pr.ShadingMaterial(**smat)

        # Handle opening_multipliers nested dataclass
        opening_multipliers_data = layer.get("opening_multipliers", {})
        if isinstance(opening_multipliers_data, dict):
            opening_multipliers = OpeningMultipliers(**opening_multipliers_data)
        else:
            opening_multipliers = OpeningMultipliers()

        layer_instances.append(
            Layer(
                product_name=layer["product_name"],
                thickness_m=layer["thickness_m"],
                product_type=layer["product_type"],
                conductivity=layer["conductivity"],
                emissivity_front=layer["emissivity_front"],
                emissivity_back=layer["emissivity_back"],
                ir_transmittance=layer["ir_transmittance"],
                flipped=layer.get("flipped", False),
                spectral_data={int(k): v for k, v in layer["spectral_data"].items()},
                coated_side=layer.get("coated_side", ""),
                shading_material=shading_material,
                slat_width_m=layer.get("slat_width_m", 0.0160),
                slat_spacing_m=layer.get("slat_spacing_m", 0.0120),
                slat_thickness_m=layer.get("slat_thickness_m", 0.0006),
                slat_curve_m=layer.get("slat_curve_m", 0.0),
                slat_angle_deg=layer.get("slat_angle_deg", 90.0),
                slat_conductivity=layer.get("slat_conductivity", 160.00),
                nslats=layer.get("nslats", 1),
                opening_multipliers=opening_multipliers,
                shading_xml=layer.get("shading_xml", ""),
            )
        )
    gaps = data.pop("gaps")
    gap_instances = []
    for gap in gaps:
        gases = gap.pop("gas")
        gas_instances = [Gas(gas=gas["gas"], ratio=gas["ratio"]) for gas in gases]
        gap_instances.append(
            Gap(
                gas=gas_instances,
                thickness_m=gap["thickness_m"],
                **{k: v for k, v in gap.items() if k not in ["gas", "thickness_m"]},
            )
        )
    return GlazingSystem(layers=layer_instances, gaps=gap_instances, **data)


def create_pwc_gaps(gaps: list[Gap]):
    """Create a list of pwc gaps from a list of gaps."""
    pwc_gaps = []
    for gap in gaps:
        _gas = pwc.create_gas(
            [[g.ratio, getattr(pwc.PredefinedGasType, g.gas.upper())] for g in gap.gas]
        )
        _gap = pwc.Layers.gap(gas=_gas, thickness=gap.thickness_m)
        pwc_gaps.append(_gap)
    return pwc_gaps


def _parse_input_source(
    input_source: Path | str | bytes,
) -> pwc.ProductData:
    """Parses various input types to pwc.ProductData."""
    if isinstance(input_source, (str, Path)):
        path = Path(input_source)
        if not path.exists():
            raise FileNotFoundError(f"{input_source} does not exist.")
        if path.suffix == ".json":
            return pwc.parse_json_file(str(path))
        elif path.suffix == ".xml":
            product_data = pwc.parse_bsdf_xml_file(str(path))
            if product_data.product_type == "" or product_data.product_type is None:
                product_data.product_type = SHADING
            return product_data
        else:
            return pwc.parse_optics_file(str(path))
    elif isinstance(input_source, bytes):
        try:
            return pwc.parse_json(input_source)
        except json.JSONDecodeError:
            product_data = pwc.parse_bsdf_xml_string(input_source.decode())
            if product_data.product_type == "" or product_data.product_type is None:
                product_data.product_type = SHADING
            return product_data


def _apply_opening_properties(
    layer_obj: Layer,
    layer_def_openings: OpeningDefinitions | None,
    gaps_list: list[Gap],
    layer_idx: int,
    total_layers_in_definition: int,
):
    """Calculates and sets opening multipliers on the layer object."""
    if layer_def_openings is None:
        return  # No opening definitions to apply

    gap_thickness = 0.0127
    if total_layers_in_definition <= 1:
        pass
    elif layer_idx == 0:
        if gaps_list:
            gap_thickness = gaps_list[0].thickness_m
    elif layer_idx == total_layers_in_definition - 1:
        if gaps_list:
            gap_thickness = gaps_list[layer_idx - 1].thickness_m
    else:  # Middle layer
        if gaps_list and len(gaps_list) > layer_idx:
            gap_thickness = min(
                gaps_list[layer_idx - 1].thickness_m, gaps_list[layer_idx].thickness_m
            )  #

    if gap_thickness > 0:
        layer_obj.opening_multipliers.top = min(
            1.0, layer_def_openings.top_m / gap_thickness
        )
        layer_obj.opening_multipliers.bottom = min(
            1.0, layer_def_openings.bottom_m / gap_thickness
        )
        layer_obj.opening_multipliers.left = min(
            1.0, layer_def_openings.left_m / gap_thickness
        )
        layer_obj.opening_multipliers.right = min(
            1.0, layer_def_openings.right_m / gap_thickness
        )
        layer_obj.opening_multipliers.front = min(
            1.0, layer_def_openings.front_multiplier
        )


def _process_blind_definition_to_bsdf(
    defin: LayerInput,
    product_data: pwc.ProductData,
    layer: Layer,
    nproc: int,
    nsamp: int,
) -> tuple[pwc.ProductData, dict]:
    """
    Handles the logic for generating BSDF for blinds if defined by geometry/material.
    Returns the pwc.ProductData of the generated BSDF and a dict of blind properties.
    """
    if isinstance(defin.input_source, bytes):
        data = json.loads(defin.input_source.decode())
    else:
        with open(defin.input_source, "r") as f:
            data = json.load(f)
    dual_band_values = data["composition"][0]["child_product"]["spectral_data"][
        "dual_band_values"
    ]
    rf_sol_diff = dual_band_values["Rf_sol_diffuse"]
    rf_sol_spec = dual_band_values["Rf_sol_specular"]
    rf_vis_diff = dual_band_values["Rf_vis_diffuse"]
    rf_vis_spec = dual_band_values["Rf_vis_specular"]
    geometry_def = product_data.geometry
    material_def = product_data.material_definition
    slat_spacing_m = geometry_def.slat_spacing * M_PER_MM
    tir = material_def.ir_transmittance
    nslats = int(1 / slat_spacing_m) if slat_spacing_m > 0 else 1
    slat_depth_m = geometry_def.slat_width * M_PER_MM
    slat_curvature_m = geometry_def.slat_curvature * M_PER_MM
    emis_front = material_def.emissivity_front  #
    emis_back = material_def.emissivity_back  #
    layer_name = product_data.product_name
    layer_thickness_m = slat_depth_m * math.cos(math.radians(defin.slat_angle_deg))
    if tir != 0:
        raise ValueError("tir not zero")
    if emis_front != emis_back:
        raise ValueError("front and back emissivity not the same")
    rf_ir = 1 - tir - emis_front

    xml = generate_blinds_xml(
        slat_depth_m,
        nslats,
        defin.slat_angle_deg,
        slat_curvature_m,
        layer_thickness_m,
        layer_name,
        rf_sol_diff,
        rf_sol_spec,
        rf_vis_diff,
        rf_vis_spec,
        rf_ir,
        nproc,
        nsamp,
    )  # meters
    actual_product_data = pwc.parse_bsdf_xml_string(xml)
    layer.thickness_m = layer_thickness_m
    layer.conductivity = product_data.material_definition.conductivity
    layer.emissivity_front = emis_front
    layer.emissivity_back = emis_back
    layer.ir_transmittance = tir
    layer.product_type = "blinds"
    layer.slat_width_m = slat_depth_m
    layer.slat_spacing_m = slat_spacing_m
    layer.slat_thickness_m = product_data.material_definition.thickness * M_PER_MM
    layer.slat_conductivity = product_data.material_definition.conductivity
    layer.slat_curve_m = slat_curvature_m
    layer.flipped = defin.flipped
    layer.shading_material = pr.ShadingMaterial(rf_vis_diff, rf_vis_spec, 0)
    layer.nslats = nslats
    layer.slat_angle_deg = defin.slat_angle_deg

    return actual_product_data


def create_glazing_system(
    name: str,
    layer_inputs: list[LayerInput],
    gaps: None | list[Gap] = None,
    nproc: int = 1,
    nsamp: int = 2000,
    mbsdf: bool = False,
) -> GlazingSystem:
    """Create a glazing system from a list of layers and gaps.

    Args:
        name: Name of the glazing system.
        layer_inputs: List of layer inputs.
        gaps: List of gaps.
        nproc: Number of processes for parallel computation.
        nsamp: Number of samples for Monte Carlo integration.
        mbsdf: Whether to use matrix BSDF.

    Returns:
        GlazingSystem object.

    Raises:
        ValueError: Invalid layer type.

    Examples:
        >>> create_glazing_system(
        ...     "test",
        ...     [
        ...         Path("glass.json"),
        ...         Path("venetian.xml"),
        ...     ],
        ... )
    """
    if gaps is None:
        gaps = [Gap([Gas("air", 1)], 0.0127) for _ in range(len(layer_inputs) - 1)]
    product_data_list: list[pwc.ProductData] = []
    layer_data: list[Layer] = []
    thickness = 0.0
    for idx, layer_inp in enumerate(layer_inputs):
        product_data = _parse_input_source(layer_inp.input_source)
        if product_data is None:
            raise ValueError("Invalid layer type")
        layer = get_layer_data(product_data)
        if product_data.product_type == SHADING:
            if product_data.product_subtype == VENETIAN:
                actual_product_data = _process_blind_definition_to_bsdf(
                    layer_inp, product_data, layer, nproc=nproc, nsamp=nsamp
                )
                layer_data.append(layer)
                product_data_list.append(actual_product_data)
                thickness += layer.thickness_m
            else:
                layer.product_type = FABRIC
                with open(layer_inp.input_source, "r") as f:
                    layer.shading_xml = f.read()
                layer_data.append(layer)
                product_data_list.append(product_data)
                thickness += layer.thickness_m
            _apply_opening_properties(
                layer, layer_inp.openings, gaps, idx, len(layer_inputs)
            )
        else:
            layer.spectral_data = {
                int(round(d.wavelength * NM_PER_MM)): (
                    d.direct_component.transmittance_front,
                    d.direct_component.reflectance_front,
                    d.direct_component.reflectance_back,
                )
                for d in product_data.measurements
            }
            layer.coated_side = product_data.coated_side
            layer_data.append(layer)
            product_data_list.append(product_data)
            thickness += layer.thickness_m

    for gap in gaps:
        thickness += gap.thickness_m

    glzsys = pwc.GlazingSystem(
        solid_layers=product_data_list,
        gap_layers=create_pwc_gaps(gaps),
        width_meters=1,
        height_meters=1,
        environment=pwc.nfrc_shgc_environments(),
        bsdf_hemisphere=pwc.BSDFHemisphere.create(pwc.BSDFBasisType.FULL),
    )

    melanopic_back_transmittace = []
    melanopic_back_reflectance = []
    if mbsdf:
        melanopic_back_transmittace, melanopic_back_reflectance = (
            generate_melanopic_bsdf(layer_data, gaps, nproc=nproc, nsamp=nsamp)
        )

    for index, data in enumerate(layer_data):
        if data.flipped:
            glzsys.flip_layer(index, True)

    solres = glzsys.optical_method_results("SOLAR")
    solsys = solres.system_results
    visres = glzsys.optical_method_results("PHOTOPIC")
    vissys = visres.system_results

    return GlazingSystem(
        name=name,
        thickness=thickness,
        layers=layer_data,
        gaps=gaps,
        solar_front_absorptance=[
            alpha.front.absorptance.angular_total for alpha in solres.layer_results
        ],
        solar_back_absorptance=[
            alpha.back.absorptance.angular_total for alpha in solres.layer_results
        ],
        visible_back_reflectance=vissys.back.reflectance.matrix,
        visible_front_reflectance=vissys.front.reflectance.matrix,
        visible_back_transmittance=vissys.back.transmittance.matrix,
        visible_front_transmittance=vissys.front.transmittance.matrix,
        solar_back_reflectance=solsys.back.reflectance.matrix,
        solar_front_reflectance=solsys.front.reflectance.matrix,
        solar_back_transmittance=solsys.back.transmittance.matrix,
        solar_front_transmittance=solsys.front.transmittance.matrix,
        melanopic_back_transmittance=melanopic_back_transmittace,
        melanopic_back_reflectance=melanopic_back_reflectance,
    )


def get_glazing_brtdfunc(name, filenames: list[str | Path]) -> pr.Primitive:
    nlayers = len(filenames)
    if nlayers > 3 or nlayers == 0:
        raise ValueError(f"Expected 1 to 3 layers, got {nlayers}.")
    panes_rgb: list[PaneRGB] = []
    for filename in filenames:
        fpath = Path(filename)
        if fpath.suffix == ".json":
            product_data = pwc.parse_json_files(str(filename))
        else:
            product_data = pwc.parse_optics_file(str(filename))
        rgb = get_layer_rgb(product_data)
        panes_rgb.append(rgb)
    return get_glazing_primitive(name, panes_rgb)


def get_pane_rgb(spectral_data: dict, coated_side: str) -> PaneRGB:
    photopic_wvl = range(380, 781, 10)
    # Filter wavelengths to only include those present in spectral_data
    available_wvl = [w for w in photopic_wvl if w in spectral_data]
    if not available_wvl:
        raise ValueError("No spectral data available in the photopic range (380-780nm)")

    tvf = [spectral_data[w][0] for w in available_wvl]
    rvf = [spectral_data[w][1] for w in available_wvl]
    rvb = [spectral_data[w][2] for w in available_wvl]
    # Use the actual wavelength range from available data
    min_wvl = min(available_wvl)
    max_wvl = max(available_wvl)
    tf_x, tf_y, tf_z = pr.spec_xyz(tvf, min_wvl, max_wvl)
    rf_x, rf_y, rf_z = pr.spec_xyz(rvf, min_wvl, max_wvl)
    rb_x, rb_y, rb_z = pr.spec_xyz(rvb, min_wvl, max_wvl)
    tf_rgb = pr.xyz_rgb(tf_x, tf_y, tf_z)
    rf_rgb = pr.xyz_rgb(rf_x, rf_y, rf_z)
    rb_rgb = pr.xyz_rgb(rb_x, rb_y, rb_z)
    if coated_side == "front":
        coated_rgb = rf_rgb
        glass_rgb = rb_rgb
    else:
        coated_rgb = rb_rgb
        glass_rgb = rf_rgb
    return PaneRGB(coated_rgb, glass_rgb, tf_rgb, coated_side)


def get_layer_rgb(layer: pwc.ProductData) -> PaneRGB:
    """Get the RGB values for a pane layer."""
    photopic_wvl = range(380, 781, 10)
    if isinstance(layer.measurements, pwc.DualBandBSDF):
        raise ValueError("Dual band measurements not supported.")
    hemi = {
        int(round(d.wavelength * NM_PER_MM)): (
            d.direct_component.transmittance_front,
            d.direct_component.transmittance_back,
            d.direct_component.reflectance_front,
            d.direct_component.reflectance_back,
        )
        for d in layer.measurements
    }
    # Filter wavelengths to only include those present in hemi data
    available_wvl = [w for w in photopic_wvl if w in hemi]
    if not available_wvl:
        raise ValueError("No spectral data available in the photopic range (380-780nm)")

    tvf = [hemi[w][0] for w in available_wvl]
    rvf = [hemi[w][2] for w in available_wvl]
    rvb = [hemi[w][3] for w in available_wvl]
    # Use the actual wavelength range from available data
    min_wvl = min(available_wvl)
    max_wvl = max(available_wvl)
    tf_x, tf_y, tf_z = pr.spec_xyz(tvf, min_wvl, max_wvl)
    rf_x, rf_y, rf_z = pr.spec_xyz(rvf, min_wvl, max_wvl)
    rb_x, rb_y, rb_z = pr.spec_xyz(rvb, min_wvl, max_wvl)
    tf_rgb = pr.xyz_rgb(tf_x, tf_y, tf_z)
    rf_rgb = pr.xyz_rgb(rf_x, rf_y, rf_z)
    rb_rgb = pr.xyz_rgb(rb_x, rb_y, rb_z)
    if layer.coated_side == "front":
        coated_rgb = rf_rgb
        glass_rgb = rb_rgb
    else:
        coated_rgb = rb_rgb
        glass_rgb = rf_rgb
    return PaneRGB(coated_rgb, glass_rgb, tf_rgb, layer.coated_side)


def get_glazing_primitive(name: str, panes: list[PaneRGB]) -> pr.Primitive:
    """Generate a BRTDfunc to represent a glazing system."""
    if len(panes) == 1:
        str_arg = [
            "sr_clear_r",
            "sr_clear_g",
            "sr_clear_b",
            "st_clear_r",
            "st_clear_g",
            "st_clear_b",
            "0",
            "0",
            "0",
            "glaze1.cal",
        ]
        coated_real = 1 if panes[0].coated_side == "front" else -1
        real_arg = [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            coated_real,
            *[round(i, 3) for i in panes[0].glass_rgb],
            *[round(i, 3) for i in panes[0].coated_rgb],
            *[round(i, 3) for i in panes[0].trans_rgb],
        ]
    elif len(panes) == 2:
        s12t_r, s12t_g, s12t_b = panes[0].trans_rgb
        s34t_r, s34t_g, s34t_b = panes[1].trans_rgb
        if panes[0].coated_side == "back":
            s2r_r, s2r_g, s2r_b = panes[0].coated_rgb
            s1r_r, s1r_g, s1r_b = panes[0].glass_rgb
        else:  # front or neither side coated
            s2r_r, s2r_g, s2r_b = panes[0].glass_rgb
            s1r_r, s1r_g, s1r_b = panes[0].coated_rgb
        if panes[1].coated_side == "back":
            s4r_r, s4r_g, s4r_b = panes[1].coated_rgb
            s3r_r, s3r_g, s3r_b = panes[1].glass_rgb
        else:  # front or neither side coated
            s4r_r, s4r_g, s4r_b = panes[1].glass_rgb
            s3r_r, s3r_g, s3r_b = panes[1].coated_rgb
        str_arg = [
            (
                f"if(Rdot,cr(fr({s4r_r:.3f}),ft({s34t_r:.3f}),fr({s2r_r:.3f})),"
                f"cr(fr({s1r_r:.3f}),ft({s12t_r:.3f}),fr({s3r_r:.3f})))"
            ),
            (
                f"if(Rdot,cr(fr({s4r_g:.3f}),ft({s34t_g:.3f}),fr({s2r_g:.3f})),"
                f"cr(fr({s1r_g:.3f}),ft({s12t_g:.3f}),fr({s3r_g:.3f})))"
            ),
            (
                f"if(Rdot,cr(fr({s4r_b:.3f}),ft({s34t_b:.3f}),fr({s2r_b:.3f})),"
                f"cr(fr({s1r_b:.3f}),ft({s12t_b:.3f}),fr({s3r_b:.3f})))"
            ),
            f"ft({s34t_r:.3f})*ft({s12t_r:.3f})",
            f"ft({s34t_g:.3f})*ft({s12t_g:.3f})",
            f"ft({s34t_b:.3f})*ft({s12t_b:.3f})",
            "0",
            "0",
            "0",
            "glaze2.cal",
        ]
        real_arg = [0, 0, 0, 0, 0, 0, 0, 0, 0]
    elif len(panes) == 3:
        str_arg = [
            "sr_red",
            "sr_grn",
            "sr_blu",
            "st_red",
            "st_grn",
            "st_blu",
            "0",
            "0",
            "0",
            "glass3.cal",
        ]
        if panes[2].coated_side == "back":
            rbspc, gbspc, bbspc = panes[2].coated_rgb
        else:
            rbspc, gbspc, bbspc = panes[2].glass_rgb
        if panes[0].coated_side == "back":
            rfspc, gfspc, bfspc = panes[0].glass_rgb
        else:
            rfspc, gfspc, bfspc = panes[0].coated_rgb
        rtspc = panes[0].trans_rgb[0] * panes[1].trans_rgb[0] * panes[2].trans_rgb[0]
        gtspc = panes[0].trans_rgb[1] * panes[1].trans_rgb[1] * panes[2].trans_rgb[1]
        btspc = panes[0].trans_rgb[2] * panes[1].trans_rgb[2] * panes[2].trans_rgb[2]
        real_arg = [
            18,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            rfspc,
            gfspc,
            bfspc,
            rbspc,
            gbspc,
            bbspc,
            rtspc,
            gtspc,
            btspc,
        ]
    else:
        raise ValueError("At most three pane supported")
    return pr.Primitive("void", "BRTDfunc", name, str_arg, real_arg)


def get_layer_groups(layers: Sequence[Layer]) -> list:
    if not layers or not isinstance(layers[0], Layer):
        raise ValueError()

    # Group consecutive glazing layers
    grouped_system: list[tuple[str, int]] = []
    current_group = 0
    current_type = ""

    for layer in layers:
        if layer.product_type == "glazing":
            if current_type == "glazing":
                current_group += 1
            else:
                if current_group:
                    grouped_system.append((current_type, current_group))
                current_group = 1
                current_type = "glazing"
        else:
            if current_group:
                grouped_system.append((current_type, current_group))
            grouped_system.append((layer.product_type, 1))
            current_group = 0
            current_type = ""

    # Add the last group if it exists
    if current_group:
        grouped_system.append((current_type, current_group))

    return grouped_system


def get_proxy_geometry(window: Polygon, gs: GlazingSystem) -> list[bytes]:
    FEPS = 1e-5
    layer_groups = get_layer_groups(gs.layers)
    window_vertices = window.vertices
    primitives: list[bytes] = []
    current_index = 0
    for group in layer_groups:
        if group[0] == "glazing":
            glazing_layer_count = group[1]
            # Get BRTDFunc
            mat_name = f"mat_{gs.name}_glazing_{current_index}"
            geom_name = f"{gs.name}_glazing_{current_index}"
            rgb = [
                get_pane_rgb(layer.spectral_data, layer.coated_side)
                for layer in gs.layers[
                    current_index : current_index + glazing_layer_count
                ]
            ]
            mat: pr.Primitive = get_glazing_primitive(mat_name, rgb)
            geom = polygon_primitive(window, mat_name, geom_name)
            primitives.append(mat.bytes)
            primitives.append(geom.bytes)
            current_index += glazing_layer_count
        elif group[0] == FABRIC:
            mat_name = f"mat_{gs.name}_fabric_{current_index}"
            geom_name = f"{gs.name}_fabric_{current_index}"
            xml_name = f"{gs.name}_fabric_{current_index}.xml"
            # Get aBSDF primitive
            with open(xml_name, "w") as f:
                f.write(gs.layers[current_index].shading_xml)
            mat: pr.Primitive = pr.Primitive(
                "void", "aBSDF", mat_name, [xml_name, "0", "0", "1", "."], []
            )
            geom = polygon_primitive(window, mat_name, geom_name)
            primitives.append(mat.bytes)
            primitives.append(geom.bytes)
            current_index += 1
        elif group[0] == "blinds":
            # NOTE: For blinds only
            zdiff1 = abs(window_vertices[1][2] - window_vertices[0][2])
            zdiff2 = abs(window_vertices[2][2] - window_vertices[1][2])
            dim1 = np.linalg.norm(window_vertices[1] - window_vertices[0])
            dim2 = np.linalg.norm(window_vertices[2] - window_vertices[1])
            if (dim1 <= FEPS) | (dim2 <= FEPS):
                print("One of the sides of the window polygon is zero")
            width: float = 0
            height: float = 0
            if zdiff1 <= FEPS:
                width = float(dim1)
                height = float(dim2)
            elif zdiff2 <= FEPS:
                width = float(dim2)
                height = float(dim1)
            else:
                print("Error: Skewed window not supported: ")
                print(window_vertices)
            blinds_layer = gs.layers[current_index]
            geom = pr.BlindsGeometry(
                depth=blinds_layer.slat_width_m,
                width=width,
                height=height,
                nslats=blinds_layer.nslats,
                angle=blinds_layer.slat_angle_deg,
                rcurv=blinds_layer.slat_curve_m,
            )
            blinds: bytes = pr.generate_blinds(blinds_layer.shading_material, geom)
            xmin, xmax, ymin, ymax, zmin, zmax = pr.getbbox(blinds)
            blinds_normal = np.array((1, 0, 0))
            rotatez_angle = angle_between(blinds_normal, window.normal, degree=True)
            rotated_blinds: bytes = pr.Xform(blinds).rotatez(rotatez_angle)()
            xmin, xmax, ymin, ymax, zmin, zmax = pr.getbbox(rotated_blinds)
            rotated_blinds_centroid = np.array(
                ((xmin + xmax) / 2, (ymin + ymax) / 2, (zmin + zmax) / 2)
            )
            blinds_at_window = pr.Xform(rotated_blinds).translate(
                *(window.centroid - rotated_blinds_centroid)
            )()
            primitives.append(
                pr.Xform(blinds_at_window).translate(
                    *(window.normal * gs.layers[current_index].thickness_m)
                )()
            )
            current_index += 1
    return primitives


def get_spectral_multi_layer_optics(
    glazings: Sequence[Layer], name: str = "unnamed"
) -> bytes:
    """Generate spectral multi-layer optics data from glazing layers."""
    # Convert Layer objects to GlazingLayerData objects
    glazing_layers = []
    for layer in glazings:
        # Extract spectral points from layer spectral_data
        spectral_points = []
        for wavelength_nm, (t, rf, rb) in layer.spectral_data.items():
            spectral_points.append(
                pr.SpectralPoint(wavelength_nm=wavelength_nm, rf=rf, rb=rb, t=t)
            )

        # Sort by wavelength
        spectral_points.sort(key=lambda x: x.wavelength_nm)

        # Determine glazing type based on layer properties
        glazing_type = pr.GlazingType.monolithic
        if layer.coated_side:
            glazing_type = pr.GlazingType.coated

        glazing_layer = pr.GlazingLayerData(
            name=layer.product_name,
            glazing_type=glazing_type,
            thickness_m=layer.thickness_m,
            spectral_points=spectral_points,
        )
        glazing_layers.append(glazing_layer)

    # this call also generates .dat files in the current working directory
    return pr.genglaze_data(glazing_layers, prefix=name)


def generate_coplanar_bsdf(
    layers: list[Layer],
    gaps: list[Gap],
    outspec: pr.genbsdf.OutSpec = "y",
    nproc: int = 1,
    nspec: int = 3,
    nsamp: int = 2000,
):
    """Generate coplanar BSDF for the given layers and gaps."""
    original_dir = os.getcwd()
    with tempfile.TemporaryDirectory() as tmpdir:
        try:
            os.chdir(tmpdir)
            tmpdir_path = Path(tmpdir)
            layer_groups = get_layer_groups(layers)
            primitives = []
            width = 1
            height = 1
            total_thickness = sum(gap.thickness_m for gap in gaps)
            total_thickness += sum(layer.thickness_m for layer in layers)

            # Start from z=0 (innermost) and work outward (negative z)
            current_z = 0
            current_layer_idx = len(layers) - 1  # Start from last layer

            # Check if we have blinds for sampling box calculation
            has_blinds = any(group[0] == "blinds" for group in layer_groups)
            blinds_spacing = 0.01  # default

            for group_idx, group in enumerate(reversed(layer_groups)):
                # Create polygon at current z position
                polygon = Polygon(
                    [
                        np.array([0, 0, current_z]),
                        np.array([0, height, current_z]),
                        np.array([width, height, current_z]),
                        np.array([width, 0, current_z]),
                    ]
                )

                if group[0] == "glazing":
                    glazing_layer_count = group[1]
                    name = f"glazing_{group_idx}"

                    # Generate spectral data for glazing layers
                    glazing_layers = layers[
                        current_layer_idx - glazing_layer_count + 1 : current_layer_idx
                        + 1
                    ]
                    glazing_material_bytes = get_spectral_multi_layer_optics(
                        glazing_layers, name=name
                    )
                    glazing_material_primitives = parse_primitive(
                        glazing_material_bytes
                    )
                    glazing_material_name = glazing_material_primitives[-1].identifier

                    # Write spectral data to file in tmpdir
                    mat_file = tmpdir_path / f"{name}.dat"
                    with open(mat_file, "wb") as f:
                        f.write(glazing_material_bytes)

                    # Create geometry primitive
                    geom = polygon_primitive(polygon, glazing_material_name, name)
                    primitives.append(glazing_material_bytes)
                    primitives.append(b"")
                    primitives.append(geom.bytes)

                    current_layer_idx -= glazing_layer_count

                    # Move outward by gap thickness
                    gap_idx = (
                        len(layer_groups) - 1 - group_idx - 1
                    )  # Corresponding gap index
                    if gap_idx >= 0 and gap_idx < len(gaps):
                        current_z -= gaps[gap_idx].thickness_m

                elif group[0] == FABRIC:
                    fabric_layer = layers[current_layer_idx]

                    mat_name = f"mat_fabric_{group_idx}"
                    geom_name = f"fabric_{group_idx}"
                    xml_name = tmpdir_path / f"fabric_{group_idx}.xml"

                    # Write fabric XML to file in tmpdir
                    with open(xml_name, "w") as f:
                        f.write(fabric_layer.shading_xml)

                    # Create aBSDF primitive
                    mat = pr.Primitive(
                        "void",
                        "aBSDF",
                        mat_name,
                        [str(xml_name), "0", "0", "1", "."],
                        [],
                    )
                    geom = polygon_primitive(polygon, mat_name, geom_name)
                    primitives.extend([mat.bytes, geom.bytes])

                    current_layer_idx -= 1

                    # Move outward by gap thickness
                    gap_idx = (
                        len(layer_groups) - 1 - group_idx - 1
                    )  # Corresponding gap index
                    if gap_idx >= 0 and gap_idx < len(gaps):
                        current_z -= gaps[gap_idx].thickness_m

                elif group[0] == "blinds":
                    blinds_layer = layers[current_layer_idx]
                    blinds_spacing = blinds_layer.slat_spacing_m

                    geom_spec = pr.BlindsGeometry(
                        depth=blinds_layer.slat_width_m,
                        width=width,
                        height=height,
                        nslats=blinds_layer.nslats,
                        angle=blinds_layer.slat_angle_deg,
                        rcurv=blinds_layer.slat_curve_m,
                    )

                    # Generate blinds geometry
                    # 0 1 0 1 -thickness 0
                    blinds = pr.generate_blinds_for_bsdf(
                        blinds_layer.shading_material, geom_spec
                    )

                    # Position blinds at polygon location
                    xmin, xmax, ymin, ymax, zmin, zmax = pr.ot.getbbox(blinds)
                    thickness = zmax - zmin
                    blinds_centroid = np.array(
                        ((xmin + xmax) / 2, (ymin + ymax) / 2, (zmin + zmax) / 2)
                    )
                    translate_vector = polygon.centroid - blinds_centroid
                    translate_vector_x = translate_vector[0]
                    translate_vector_y = translate_vector[1]
                    translate_vector_z = translate_vector[2] - thickness / 2
                    positioned_blinds = pr.Xform(blinds).translate(
                        translate_vector_x,
                        translate_vector_y,
                        translate_vector_z,
                    )()
                    primitives.append(positioned_blinds)

                    current_layer_idx -= 1

                    # Move outward by gap thickness minus half blinds depth (for blinds space)
                    gap_idx = (
                        len(layer_groups) - 1 - group_idx - 1
                    )  # Corresponding gap index
                    if gap_idx >= 0 and gap_idx < len(gaps):
                        gap_thickness = gaps[gap_idx].thickness_m
                        current_z -= gap_thickness
                        current_z -= thickness

            device_file = tmpdir_path / "device.rad"
            with open(device_file, "wb") as f:
                for primitive in primitives:
                    f.write(primitive)

            box_height = blinds_spacing if has_blinds else 0.01
            sampling_box = pr.genbsdf.SamplingBox(
                xmin=0.5,
                xmax=0.51,
                ymin=0.5,
                ymax=0.5 + box_height,
                zmin=-total_thickness,
                zmax=0,
            )

            bsdf_result = pr.generate_bsdf(
                str(device_file),
                basis="kf",
                outspec=outspec,
                dim=sampling_box,
                nproc=nproc,
                nspec=nspec,
                nsamp=nsamp,
            )
        finally:
            os.chdir(original_dir)
    return bsdf_result


def generate_melanopic_bsdf(
    layers: list[Layer],
    gaps: list[Gap],
    nproc: int = 1,
    nsamp: int = 2000,
    nspec: int = 20,
) -> tuple[list[list], list[list]]:
    """Generate melanopic BSDF for the glazing system."""
    data = generate_coplanar_bsdf(
        layers, gaps, outspec="m", nspec=nspec, nproc=nproc, nsamp=nsamp
    )
    back_transmittance = [
        list(map(float, row.split()))
        for row in data.back.transmittance.decode().splitlines()
    ]
    back_reflectance = [
        list(map(float, row.split()))
        for row in data.back.reflectance.decode().splitlines()
    ]
    return back_transmittance, back_reflectance



from dataclasses import dataclass

import numpy as np
import pyradiance as pr
from frads.geom import Polygon, polygon_primitive


@dataclass
class WindowSurface:
    polygon: Polygon
    primitive: pr.Primitive


@dataclass
class Surface:
    base: Polygon
    base_primitive: pr.Primitive
    polygons: list[Polygon]
    windows: list[WindowSurface]
    modifier: str
    identifier: str
    primitives: list[pr.Primitive]

    def move_window(self, distance: float) -> None:
        """Move windows in its normal direction."""
        direction = self.base.normal * distance
        new_windows = []
        for window in self.windows:
            new_polygon = window.polygon.move(direction)
            new_primitive = polygon_primitive(
                polygon=new_polygon,
                modifier=window.primitive.modifier,
                identifier=window.primitive.identifier,
            )
            new_windows.append(WindowSurface(new_polygon, new_primitive))
        self.windows = new_windows

    def rotate_z(self, radians):
        """Rotate the surface counter clock-wise."""
        center = np.zeros(3)
        zaxis = np.array((0, 0, 1))
        new_base = self.base.rotate(center, zaxis, radians)
        new_base_primitive = polygon_primitive(
            polygon=new_base,
            modifier=self.base_primitive.modifier,
            identifier=self.base_primitive.identifier,
        )
        new_polygons = [plg.rotate(center, zaxis, radians) for plg in self.polygons]
        new_primitives = []
        for idx, polygon in enumerate(new_polygons):
            new_primitives.append(
                polygon_primitive(
                    polygon=polygon,
                    modifier=self.primitives[idx].modifier,
                    identifier=self.primitives[idx].identifier,
                )
            )
        new_windows = []
        for window in self.windows:
            new_polygon = window.polygon.rotate(center, zaxis, radians)
            new_primitive = polygon_primitive(
                polygon=new_polygon,
                modifier=window.primitive.modifier,
                identifier=window.primitive.identifier,
            )
            new_windows.append(WindowSurface(new_polygon, new_primitive))

        self.polygons = new_polygons
        self.primitives = new_primitives
        self.windows = new_windows
        self.base = new_base
        self.base_primitive = new_base_primitive


@dataclass
class Room:
    floor: Surface
    ceiling: Surface
    swall: Surface
    ewall: Surface
    nwall: Surface
    wwall: Surface
    materials: list[pr.Primitive]

    def primitives(self) -> list[pr.Primitive]:
        return [
            *self.floor.primitives,
            *self.ceiling.primitives,
            *self.swall.primitives,
            *self.ewall.primitives,
            *self.nwall.primitives,
            *self.wwall.primitives,
        ]

    def window_primitives(self) -> list[pr.Primitive]:
        return [
            *[srf.primitive for srf in self.ceiling.windows],
            *[srf.primitive for srf in self.swall.windows],
            *[srf.primitive for srf in self.ewall.windows],
            *[srf.primitive for srf in self.nwall.windows],
            *[srf.primitive for srf in self.wwall.windows],
        ]

    def model_dump(self) -> dict:
        model = {}
        model["materials"] = {"bytes": b" ".join(p.bytes for p in self.materials)}
        model["scene"] = {"bytes": b" ".join(p.bytes for p in self.primitives())}
        model["windows"] = {}
        for primitive in self.window_primitives():
            model["windows"][primitive.identifier] = {"bytes": primitive.bytes}
        model["surfaces"] = {
            "floor": {"primitives": [self.floor.base_primitive]},
            "ceiling": {"primitives": [self.ceiling.base_primitive]},
            "swall": {"primitives": [self.swall.base_primitive]},
            "ewall": {"primitives": [self.ewall.base_primitive]},
            "nwall": {"primitives": [self.nwall.base_primitive]},
            "wwall": {"primitives": [self.wwall.base_primitive]},
        }
        return model

    def rotate_z(self, radians):
        """Rotate the room counter clock-wise."""
        self.floor.rotate_z(radians)
        self.ceiling.rotate_z(radians)
        self.swall.rotate_z(radians)
        self.ewall.rotate_z(radians)
        self.wwall.rotate_z(radians)
        self.nwall.rotate_z(radians)

    def validate(self) -> None:
        """Validate the room model."""
        material_names = [p.identifier for p in self.materials]
        for prim in self.primitives():
            if prim.modifier not in material_names:
                raise ValueError(
                    f"Unknown modifier {prim.modifier} in {prim.identifier}"
                )


def neutral_plastic_prim(
    mod: str, ident: str, refl: float, spec: float, rough: float
) -> pr.Primitive:
    """
    Generate a neutral color plastic material.

    Args:
        mod: modifier to the primitive
        ident: identifier to the primitive
        refl: measured reflectance (0.0 - 1.0)
        spec: material specularity (0.0 - 1.0)
        rough: material roughness (0.0 - 1.0)

    Returns:
        A material primtive
    """
    err_msg = "reflectance, speculariy, and roughness have to be 0-1"
    assert all(0 <= i <= 1 for i in [spec, refl, rough]), err_msg
    real_args = [refl, refl, refl, spec, rough]
    return pr.Primitive(mod, "plastic", ident, [], real_args)


def glass_prim(
    mod: str, ident: str, tr: float, tg: float, tb: float, refrac: float = 1.52
) -> pr.Primitive:
    """Generate a glass material.

    Args:
        mod: modifier to the primitive
        ident: identifier to the primtive
        tr: Transmissivity in red channel (0.0 - 1.0)
        tg: Transmissivity in green channel (0.0 - 1.0)
        tb: Transmissivity in blue channel (0.0 - 1.0)
        refrac: refraction index (default=1.52)
    Returns:
        material primtive (dict)

    """
    tmsv_red = tr * 1.08981
    tmsv_green = tg * 1.08981
    tmsv_blue = tb * 1.08981
    real_args = [tmsv_red, tmsv_green, tmsv_blue, refrac]
    return pr.Primitive(mod, "glass", ident, [], real_args)


def material_lib() -> dict[str, pr.Primitive]:
    """Generate a list of generic material primitives."""
    tmis = 0.6 * 1.08981
    return {
        "neutral_lambertian_0.2": neutral_plastic_prim(
            "void", "neutral_lambertian_0.2", 0.2, 0, 0
        ),
        "neutral_lambertian_0.5": neutral_plastic_prim(
            "void", "neutral_lambertian_0.5", 0.5, 0, 0
        ),
        "neutral_lambertian_0.7": neutral_plastic_prim(
            "void", "neutral_lambertian_0.7", 0.7, 0, 0
        ),
        "glass_60": glass_prim("void", "glass_60", tmis, tmis, tmis),
    }


def thicken(base, thickness) -> list[Polygon]:
    """Thicken the surface."""
    direction = base.normal * thickness
    polygons = base.extrude(direction)
    # Remove duplicates.
    counts = [polygons.count(plg) for plg in polygons]
    polygons = [plg for plg, cnt in zip(polygons, counts) if cnt == 1]
    return polygons


def make_window(
    name: str,
    base: Polygon,
    vertices: np.ndarray,
    upvec: np.ndarray,
    rightvec: np.ndarray,
    dist_left: float,
    dist_bot: float,
    width: float,
    height: float,
) -> tuple[Polygon, WindowSurface]:
    """Make one or more window and punch a hole.

    -----------------
    |      __       |
    |     |  |      |
    |__a__|__|      |
    |     |         |
    |     |b        |
    -----------------

    a: dist_left
    b: dist_bot

    Args:
        base: The base polygon.
        vertices: The vertices of the original base polygon.
        vec1: The unit vector of the first edge of the original base polygon.
        vec2: The unit vector of the second edge of the original base polygon.
        dist_left: The distance from the left edge of the base polygon.
        dist_bot: The distance from the bottom edge of the base polygon.
        width: The width of the window.
        height: The height of the window.
    """
    win_pt1 = vertices[0] + upvec * dist_bot + rightvec * dist_left
    win_pt2 = win_pt1 + upvec * height
    win_pt3 = win_pt1 + rightvec * width
    window_polygon = Polygon.rectangle3pts(win_pt3, win_pt1, win_pt2)
    new_base = base - window_polygon
    return new_base, WindowSurface(
        polygon=window_polygon,
        primitive=polygon_primitive(
            polygon=window_polygon,
            modifier="glass_60",
            identifier=name,
        ),
    )


def make_window_wwr(base: Polygon, wwr: float) -> tuple[Polygon, WindowSurface]:
    """
    Make a single window and punch a hole based on window-to-wall ratio.

    """
    window_polygon = base.scale(np.array((wwr, wwr, wwr)), base.centroid)
    base = base - window_polygon
    return base, WindowSurface(
        polygon=window_polygon,
        primitive=polygon_primitive(
            polygon=window_polygon,
            modifier="glass_60",
            identifier="void",
        ),
    )


def create_surface(
    base: Polygon,
    thickness: float = 0,
    modifier: str = "void",
    identifier: str = "void",
    wpd: None | list[list[float]] = None,
    wwr: None | float = None,
) -> Surface:
    """Create a surface with windows.

    Args:
        base: The base polygon.
        thickness: The thickness of the surface.
        modifier: The modifier of the surface.
        identifier: The identifier of the surface.
        wpd: The window position and dimension, mutex to wwr.
        wwr: The window-to-wall ratio, mutex to wpd.

    Returns:
        A Surface object.
    """
    vertices = base.vertices
    polygons = [base]
    windows = []
    base_primitive = polygon_primitive(
        polygon=base,
        modifier=modifier,
        identifier=identifier,
    )
    primitives = [
        polygon_primitive(
            polygon=base,
            modifier=modifier,
            identifier=identifier,
        )
    ]
    if wpd is not None:
        # Make a window based on window position and dimension.
        for i, pd in enumerate(wpd):
            name = f"{identifier}_window{i}"
            # vec1 = (vertices[1] - vertices[0]) / np.linalg.norm(vertices[1] - vertices[0])
            # vec2 = (vertices[2] - vertices[1]) / np.linalg.norm(vertices[2] - vertices[1])
            upvec = np.array((0, 0, 1))
            rightvec = np.array((1, 0, 0))
            base, _window = make_window(name, base, vertices, upvec, rightvec, *pd)
            windows.append(_window)
    elif wwr is not None:
        # Make a window based on window-to-wall ratio.
        base, _window = make_window_wwr(base, wwr)
        windows.append(_window)
    if thickness > 0:
        polygons = thicken(base, thickness)
        primitives = [
            polygon_primitive(
                polygon=polygon,
                modifier=modifier,
                identifier=f"{identifier}_{i}",
            )
            for i, polygon in enumerate(polygons)
        ]
    return Surface(
        base,
        base_primitive,
        polygons,
        windows,
        modifier,
        identifier,
        primitives,
    )


def create_south_facing_room(
    width: float,
    depth: float,
    floor_floor: float,
    floor_ceiling: float,
    swall_thickness: float = 0,
    wpd: None | list[list[float]] = None,
    wwr: None | float = None,
) -> Room:
    materials = list(material_lib().values())
    pt1 = np.array((0, 0, 0))
    pt2 = pt1 + np.array((0, depth, 0))
    pt3 = pt2 + np.array((width, 0, 0))
    base_floor = Polygon.rectangle3pts(pt1, pt2, pt3)
    # _, base_ceiling, base_wwall, base_nwall, base_ewall, base_swall = (
    #     base_floor.extrude(np.array((0, 0, floor_floor)))
    # )
    base_ceiling = base_floor.flip().move(np.array((0, 0, floor_ceiling)))
    base_nwall = Polygon(
        [
            np.array((width, depth, 0)),
            np.array((0, depth, 0)),
            np.array((0, depth, floor_floor)),
            np.array((width, depth, floor_floor)),
        ]
    )
    base_swall = Polygon(
        [
            np.array((0, 0, 0)),
            np.array((width, 0, 0)),
            np.array((width, 0, floor_floor)),
            np.array((0, 0, floor_floor)),
        ]
    )
    base_ewall = Polygon(
        [
            np.array((width, 0, 0)),
            np.array((width, depth, 0)),
            np.array((width, depth, floor_floor)),
            np.array((width, 0, floor_floor)),
        ]
    )
    base_wwall = Polygon(
        [
            np.array((0, 0, 0)),
            np.array((0, 0, floor_floor)),
            np.array((0, depth, floor_floor)),
            np.array((0, depth, 0)),
        ]
    )
    floor = create_surface(
        base_floor,
        modifier="neutral_lambertian_0.2",
        identifier="floor",
    )
    ceiling = create_surface(
        base_ceiling,
        modifier="neutral_lambertian_0.7",
        identifier="ceiling",
    )
    nwall = create_surface(
        base_nwall,
        modifier="neutral_lambertian_0.5",
        identifier="nwall",
    )
    ewall = create_surface(
        base_ewall,
        modifier="neutral_lambertian_0.5",
        identifier="ewall",
    )
    wwall = create_surface(
        base_wwall,
        modifier="neutral_lambertian_0.5",
        identifier="wwall",
    )
    swall = create_surface(
        base_swall,
        thickness=swall_thickness,
        modifier="neutral_lambertian_0.5",
        identifier="swall",
        wpd=wpd,
        wwr=wwr,
    )
    return Room(
        floor,
        ceiling,
        swall,
        ewall,
        nwall,
        wwall,
        materials,
    )



"""
This module contains all utility functions used throughout frads.
"""

from io import TextIOWrapper
import logging
from pathlib import Path
from random import choices
import string

import numpy as np
from pyradiance import Primitive, parse_primitive, pvaluer


logger: logging.Logger = logging.getLogger("frads.utils")


def array_hdr(array: np.ndarray, xres: int, yres: int, dtype: str = "d") -> bytes:
    """
    Call pvalue to generate a HDR image from a numpy array.
    Args:
        array: one-dimensional pixel values [[r1, g1, b1], [r2, g2, b2], ...]
        xres: x resolution
        yres: y resolution
        dtype: data type of the array. 'd' for double, 'f' for float
    Returns:
        HDR image in bytes
    """
    return pvaluer(array.tobytes(), inform=dtype, header=False, xres=xres, yres=yres)


def write_hdr(
    fname: str, array: np.ndarray, xres: int, yres: int, dtype: str = "d"
) -> None:
    """
    Write a array into a HDR image.
    Args:
        fname: output file name
        array: one-dimensional pixel values [[r1, g1, b1], [r2, g2, b2], ...]
        xres: x resolution
        yres: y resolution
        dtype: data type of the array. 'd' for double, 'f' for float
    Returns:
        None
    """
    with open(fname, "wb") as f:
        f.write(array_hdr(array, xres, yres, dtype=dtype))


def write_hdrs(
    array: np.ndarray, xres: int, yres: int, dtype: str = "d", outdir: str = "."
) -> None:
    """
    Write a series of HDR images to a file.
    Args:
        array: two-dimensional pixel values [[r1, g1, b1], [r2, g2, b2], ...]
            where each column of data represents a image.
        xres: x resolution
        yres: y resolution
        dtype: data type of the array. 'd' for double, 'f' for float
        outdir: output directory
    Returns:
        None
    """
    if not Path(outdir).exists():
        Path(outdir).mkdir(parents=True)
    # iterate through the columns of array
    for i in range(array.shape[1]):
        fname = f"{outdir}/hdr_{i}.hdr"
        write_hdr(fname, array[:, i], xres, yres, dtype=dtype)


def write_ep_rad_model(outpath: str, model: dict) -> None:
    """
    Write a epjson2rad model into a Radiance file.
    Args:
        outpath: output file path
        model: a model object
    """
    with open(outpath, "wb") as f:
        f.write(model["model"]["materials"]["bytes"])
        f.write(model["model"]["scene"]["bytes"])
        for window in model["model"]["windows"].values():
            f.write(window.bytes)


def unpack_primitives(file: str | Path | TextIOWrapper) -> list[Primitive]:
    """Open a file a to parse primitive."""
    if isinstance(file, TextIOWrapper):
        lines = file.read()
    else:
        with open(file, "r", encoding="ascii") as rdr:
            lines = rdr.read()
    return parse_primitive(lines)


def neutral_trans_prim(
    mod: str, ident: str, trans: float, refl: float, spec: float, rough: float
) -> Primitive:
    """
    Generate a neutral color plastic material.

    Args:
        mod: modifier to the primitive
        ident: identifier to the primitive
        refl: measured reflectance (0.0 - 1.0)
        spec: material specularity (0.0 - 1.0)
        rough: material roughness (0.0 - 1.0)

    Returns:
        A material primtive
    """
    color = trans + refl
    t_diff = trans / color
    tspec = 0
    err_msg = "reflectance, speculariy, and roughness have to be 0-1"
    assert all(0 <= i <= 1 for i in [spec, refl, rough]), err_msg
    real_args = [color, color, color, spec, rough, t_diff, tspec]
    return Primitive(mod, "trans", ident, [], real_args)


def color_plastic_prim(
    mod: str,
    ident: str,
    refl: float,
    red: int,
    green: int,
    blue: int,
    specu: float,
    rough: float,
) -> Primitive:
    """Generate a colored plastic material.
    Args:
        mod: modifier to the primitive
        ident: identifier to the primitive
        refl : measured reflectance (0.0 - 1.0)
        red: green; blue (int): rgb values (0 - 255)
        specu: material specularity (0.0 - 1.0)
        rough: material roughness (0.0 - 1.0)

    Returns:
        A material primtive
    """
    err_msg = "reflectance, speculariy, and roughness have to be 0-1"
    assert all(0 <= i <= 1 for i in [specu, refl, rough]), err_msg
    red_eff = 0.3
    green_eff = 0.59
    blue_eff = 0.11
    weighted = red * red_eff + green * green_eff + blue * blue_eff
    matr = round(red / weighted * refl, 3)
    matg = round(green / weighted * refl, 3)
    matb = round(blue / weighted * refl, 3)
    real_args = [matr, matg, matb, specu, rough]
    return Primitive(mod, "plastic", ident, [], real_args)


def random_string(size: int) -> str:
    """Generate random characters."""
    chars = string.ascii_uppercase + string.digits
    return "".join(choices(chars, k=size))


# def add_manikin(
#     manikin_file: str,
#     manikin_name: str,
#     zone: dict,
#     position: list[float],
#     rotation: float = 0,
# ) -> None:
#     """Add a manikin to the scene.i
#     Args:
#         manikin_file: path to the manikin file
#         manikin_name: name of the manikin
#         zone: zone as a dictionary, must have 'model' key, we assume all scene
#             data is inside the data key, and not files.
#         position: position of the manikin (x, y), where x and y are 0-1
#         rotation: rotation of the manikin in degree (0-360)
#     Returns:
#         A zone with added manikin
#     Notes:
#         Zone dictionary is modified in place.
#     """
#     zone["model"]["scene"]["bytes"] += b" "
#     zone_primitives = parse_primitive(zone["model"]["scene"]["bytes"].decode())
#     zone_polygons = [parse_polygon(p) for p in zone_primitives if p.ptype == "polygon"]
#     xmin, xmax, ymin, ymax, zmin, _ = fr.geom.get_polygon_limits(zone_polygons)
#     target = np.array(
#         [xmin + (xmax - xmin) * position[0], ymin + (ymax - ymin) * position[1], zmin]
#     )
#     with open(manikin_file) as f:
#         manikin_primitives = parse_primitive(f.read())
#     non_polygon_primitives = [p for p in manikin_primitives if p.ptype != "polygon"]
#     for primitive in non_polygon_primitives:
#         zone["model"]["scene"]["bytes"] += primitive.bytes
#     manikin_polygons = [
#         parse_polygon(p) for p in manikin_primitives if p.ptype == "polygon"
#     ]
#     xminm, xmaxm, yminm, ymaxm, zminm, _ = fr.geom.get_polygon_limits(manikin_polygons)
#     manikin_base_center = np.array([(xmaxm - xminm) / 2, (ymaxm - yminm) / 2, zminm])
#     if rotation != 0:
#         manikin_polygons = [
#             p.rotate(manikin_base_center, np.array([0, 0, 1]), np.radians(rotation))
#             for p in manikin_polygons
#         ]
#     move_vector = manikin_base_center - target
#     moved_manikin_polygons = [polygon.move(move_vector) for polygon in manikin_polygons]
#     moved_manikin = [
#         polygon_primitive(polygon, primitive.modifier, primitive.identifier)
#         for polygon, primitive in zip(moved_manikin_polygons, manikin_primitives)
#     ]
#     for primitive in moved_manikin:
#         zone["model"]["scene"]["bytes"] += primitive.bytes
#     manikin_rays = []
#     for polygon in moved_manikin_polygons:
#         manikin_rays.append([*polygon.centroid.tolist(), *polygon.normal.tolist()])
#     zone["model"]["sensors"][manikin_name] = {"data": manikin_rays}



"""
Routines for generating sky models
"""

import datetime
import logging
import math
import os
from pathlib import Path
from typing import NamedTuple, Sequence

import pyradiance as pr

logger: logging.Logger = logging.getLogger("frads.sky")


class WeaMetaData(NamedTuple):
    """Weather related meta data object.

    Attributes:
        city: City.
        country: Country.
        latitude: Latitude.
        longitude: Longitude.
        timezone: Timezone as standard meridian.
        elevation: Site elevation (m).
    """

    city: str
    country: str
    latitude: float
    longitude: float
    timezone: int
    elevation: float
    unit: int = 1

    def wea_header(self) -> str:
        """Return a .wea format header."""
        return (
            f"place {self.city}_{self.country}\n"
            f"latitude {self.latitude}\n"
            f"longitude {self.longitude}\n"
            f"time_zone {self.timezone}\n"
            f"site_elevation {self.elevation}\n"
            f"weather_data_file_units {self.unit}\n"
        )


class WeaData(NamedTuple):
    """Weather related data object.

    Attributes:
        month: Month.
        day: Day.
        hour: Hour.
        minute: Minutes.
        second: Seconds.
        hours: Times with minutes as fraction.
        dni: Direct normal irradiance (W/m2) or illuminance (lux).
        dhi: Diffuse horizontal irradiance (W/m2) or illuminance (lux).
        aod: Aeroal Optical Depth (default = 0).
        cc: Cloud cover (default = 0).
        year: default = 2000.
    """

    time: datetime.datetime
    dni: float
    dhi: float
    aod: float = 0
    cc: float = 0

    def __str__(self) -> str:
        return f"{self.time.month} {self.time.day} {self.time.hour + self.time.minute / 60} {self.dni} {self.dhi} {self.aod} {self.cc}"

    def dt_str(self) -> str:
        return f"{self.time.month:02d}{self.time.day:02d}_{self.time.hour:02d}{self.time.minute:02d}"


def parse_epw(epw_str: str) -> tuple:
    """Parse epw file and return wea header and data.

    Args:
        epw_str: String containing epw file.

    Returns:
        Tuple of meta data and wea data.
    """
    raw = epw_str.splitlines()
    epw_header = raw[0].split(",")
    content = raw[8:]
    data = []
    for li in content:
        line = li.split(",")
        year = int(line[0])
        month = int(line[1])
        day = int(line[2])
        hour = int(line[3]) - 1
        dir_norm = float(line[14])
        dif_hor = float(line[15])
        cc = float(line[19])
        aod = float(line[26])
        data.append(
            WeaData(
                datetime.datetime(year, month, day, hour, 30),
                dir_norm,
                dif_hor,
                cc,
                aod,
            )
        )
    city = epw_header[1]
    country = epw_header[3]
    latitude = float(epw_header[6])
    longitude = -1 * float(epw_header[7])
    tz = int(float(epw_header[8])) * (-15)
    elevation = float(epw_header[9].rstrip())
    meta_data = WeaMetaData(city, country, latitude, longitude, tz, elevation)
    return meta_data, data


def parse_wea(wea_str: str) -> tuple[WeaMetaData, list[WeaData]]:
    """Parse a WEA weather file.

    Args:
        wea_str: String containing the complete WEA file content.

    Returns:
        A tuple containing:
            - WeaMetaData: Location and site information
            - list[WeaData]: List of weather data entries for each time step

    Examples:
        >>> with open('weather.wea', 'r') as f:
        ...     content = f.read()
        >>> metadata, data = parse_wea(content)
        >>> print(f"Location: {metadata.city}")
        >>> print(f"Data points: {len(data)}")
    """
    lines = wea_str.splitlines()
    place = lines[0].split(" ", 1)[1]
    lat = float(lines[1].split(" ", 1)[1])
    lon = float(lines[2].split(" ", 1)[1])
    tz = int(float(lines[3].split(" ", 1)[1]))
    ele = float(lines[4].split(" ", 1)[1])
    meta_data = WeaMetaData(place, "", lat, lon, tz, ele)
    year = datetime.datetime.today().year
    data = []
    for li in lines[6:]:
        if li.strip() == "":
            continue
        line = li.split()
        month = int(line[0])
        day = int(line[1])
        hours = float(line[2])
        hour = int(hours)
        minute = int((hours - hour) * 60)
        dir_norm = float(line[3])
        dif_hor = float(line[4])
        data.append(
            WeaData(
                datetime.datetime(year, month, day, hour, minute), dir_norm, dif_hor
            )
        )
    return meta_data, data


def parse_epw_file(file: os.PathLike) -> tuple:
    """Parse an epw file using parse_epw.

    Args:
        file: Path to epw file.

    Returns:
        tuple of meta data and epw data.
    """
    with open(file, "r") as f:
        epw_str = f.read()
    return parse_epw(epw_str)


def parse_wea_file(file: os.PathLike) -> tuple:
    """Parse a wea file using parse_wea.

    Args:
        file: Path to wea file.

    Returns:
        tuple of meta data and wea data.
    """
    with open(file, "r") as f:
        wea_str = f.read()
    return parse_wea(wea_str)


def genskymtx(
    data: None | Sequence[WeaData] = None,
    meta: None | WeaMetaData = None,
    wpath: None | str | Path = None,
    onesun: bool = False,
    header: bool = True,
    average: bool = False,
    sun_only: bool = False,
    sky_only: bool = False,
    sun_file: None | str = None,
    sun_mods: None | str = None,
    daylight_hours_only: bool = False,
    sky_color: None | list[float] = None,
    ground_color: None | list[float] = None,
    rotate: None | float = None,
    outform: None | str = None,
    solar_radiance: bool = False,
    mfactor: int = 1,
) -> bytes:
    """Call gendaymtx to generate a sky/sun matrix
    Write results to out.  It takes either a .wea file path
    or wea data and metadata (defined in frads.types).
    If both are provided, .wea file path will be used.

    Args:
        data: A list of WeaData objects.
        meta: A WeaMetaData object.
        wpath: A .wea file path.
        onesun: If True, only one sun will be generated.
        header: If True, a header will be included in the output.
        average: If True, the output will be averaged.
        sun_only: If True, only sun will be generated.
        sky_only: If True, only sky will be generated.
        sun_file: A sun file path.
        sun_mods: A sun modifier.
        daylight_hours_only: If True, only daylight hours will be generated.
        sky_color: A list of sky color values.
        ground_color: A list of ground color values.
        rotate: A rotation value.
        outform: An output format.
        solar_radiance: If True, solar radiance will be generated.
        mfactor: An mfactor value.

    Returns:
        A bytes object containing the output.

    Raises:
        ValueError: An error occurs if neither a .wea path nor wea data is provided.
    """
    if wpath is None:
        if data is not None and meta is not None:
            inp = gen_wea(
                [d.time for d in data],
                [d.dni for d in data],
                [d.dhi for d in data],
                meta.latitude,
                meta.longitude,
                meta.timezone,
                elevation=meta.elevation,
                location=meta.city + meta.country,
            ).encode()
        else:
            raise ValueError("Either a .wea path or wea data is required.")
    else:
        inp = wpath
    _out = pr.gendaymtx(
        inp,
        header=header,
        average=average,
        sun_only=sun_only,
        sky_only=sky_only,
        solar_radiance=solar_radiance,
        sun_file=sun_file,
        sun_mods=sun_mods,
        daylight_hours_only=daylight_hours_only,
        sky_color=sky_color,
        ground_color=ground_color,
        rotate=rotate,
        outform=outform,
        onesun=onesun,
        mfactor=mfactor,
    )
    return _out


def gen_perez_sky(
    dt: datetime.datetime,
    latitude: float,
    longitude: float,
    timezone: int,
    year: None | int = None,
    dirnorm: None | float = None,
    diffhor: None | float = None,
    dirhor: None | float = None,
    dirnorm_illum: None | float = None,
    diffhor_illum: None | float = None,
    solar: bool = False,
    grefl: None | float = None,
    rotate: None | float = None,
) -> bytes:
    """Generate a Perez sky model using the gendaylit Radiance command.

    Creates a realistic sky luminance distribution based on the Perez all-weather
    sky model, which accounts for varying sky conditions from clear to overcast.

    Args:
        dt: Date and time for the sky calculation.
        latitude: Site latitude in degrees (positive North, negative South).
        longitude: Site longitude in degrees (positive East, negative West).
        timezone: Standard meridian for timezone in degrees.
        year: Year for the calculation (optional, uses dt.year if None).
        dirnorm: Direct normal irradiance in W/m² (optional).
        diffhor: Diffuse horizontal irradiance in W/m² (optional).
        dirhor: Direct horizontal irradiance in W/m² (optional).
        dirnorm_illum: Direct normal illuminance in lux (optional).
        diffhor_illum: Diffuse horizontal illuminance in lux (optional).
        solar: If True, include solar disk in the sky model.
        grefl: Ground reflectance (0.0-1.0, default varies by season).
        rotate: Sky rotation angle in degrees (optional).

    Returns:
        bytes: the sky primitive.
    """
    sun = pr.gendaylit(
        dt,
        latitude,
        longitude,
        timezone,
        year,
        dirnorm=dirnorm,
        diffhor=diffhor,
        dirhor=dirhor,
        dirnorm_illum=dirnorm_illum,
        diffhor_illum=diffhor_illum,
        solar=solar,
        grefl=grefl,
    )
    if rotate:
        sun = pr.xform(sun, rotatez=rotate)
    out = [pr.Primitive("skyfunc", "glow", "sglow", [], [1, 1, 1, 0]).bytes]
    out.append(pr.Primitive("sglow", "source", "sky", [], [0, 0, 1, 180]).bytes)
    out.append(pr.Primitive("sglow", "source", "ground", [], [0, 0, -1, 180]).bytes)
    return sun + b" ".join(out)


def gen_wea(
    datetimes: Sequence[datetime.datetime],
    dirnorm: Sequence[float],
    diffhor: Sequence[float],
    latitude: float,
    longitude: float,
    timezone: int,
    elevation: None | float = None,
    location: None | str = None,
) -> str:
    """Generate wea file from datetime, location, and sky."""
    if len(datetimes) != len(dirnorm) != len(diffhor):
        raise ValueError("datetimes, dirnorm, and diffhor must be the same length")
    rows = []
    if location is None or location == "":
        location = "_".join(
            [str(i) for i in [latitude, longitude, timezone, elevation]]
        )
    if elevation is None:
        elevation = 0
    rows.append(f"place {location}")
    rows.append(f"latitude {latitude}")
    rows.append(f"longitude {longitude}")
    rows.append(f"time_zone {timezone}")
    rows.append(f"site_elevation {elevation}")
    rows.append("weather_data_file_units 1")
    for dt, dni, dhi in zip(datetimes, dirnorm, diffhor):
        _hrs = dt.hour + dt.minute / 60  # middle of hour
        _row = f"{dt.month} {dt.day} {_hrs} {dni} {dhi}"
        rows.append(_row)
    return "\n".join(rows)


def solar_angle(
    lat: float,
    lon: float,
    mer: float,
    month,
    day: int,
    hour,
) -> tuple[float, float]:
    """Calculate solar altitude and azimuth angles for a given location and time.

    This is a simplified translation from the Radiance sun.c and gensky.c code.
    Computes the solar position using astronomical formulas.

    Args:
        lat: Latitude in degrees (positive for North, negative for South).
        lon: Longitude in degrees (positive for East, negative for West).
        mer: Standard meridian for the time zone in degrees.
        month: Month (1-12).
        day: Day of month (1-31).
        hour: Hour in decimal format (e.g., 13.5 for 1:30 PM).

    Returns:
        A tuple containing:
            - altitude: Solar altitude angle in radians (angle above horizon)
            - azimuth: Solar azimuth angle in radians (angle from North, clockwise)

    Examples:
        >>> # Calculate sun position for Oakland, CA on June 21st at noon
        >>> alt, az = solar_angle(37.8, -122.3, -120, 6, 21, 12.0)
        >>> print(f"Altitude: {math.degrees(alt):.1f}°")
        >>> print(f"Azimuth: {math.degrees(az):.1f}°")
    """
    latitude_r = math.radians(lat)
    longitude_r = math.radians(lon)
    s_meridian = math.radians(mer)
    mo_da = [0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334]

    julian_date = mo_da[month - 1] + day

    solar_decline = 0.4093 * math.sin((2 * math.pi / 365) * (julian_date - 81))

    solar_time = hour + (
        0.170 * math.sin((4 * math.pi / 373) * (julian_date - 80))
        - 0.129 * math.sin((2 * math.pi / 355) * (julian_date - 8))
        + (12 / math.pi) * (s_meridian - longitude_r)
    )
    altitude = math.asin(
        math.sin(latitude_r) * math.sin(solar_decline)
        - math.cos(latitude_r)
        * math.cos(solar_decline)
        * math.cos(solar_time * (math.pi / 12))
    )
    azimuth = -math.atan2(
        math.cos(solar_decline) * math.sin(solar_time * (math.pi / 12.0)),
        -math.cos(latitude_r) * math.sin(solar_time)
        - math.sin(latitude_r)
        * math.cos(solar_decline)
        * math.cos(solar_time * (math.pi / 12)),
    )

    return altitude, azimuth


def start_end_hour(data: Sequence[WeaData], sh: float, eh: float) -> Sequence[WeaData]:
    """Remove wea data entries outside of the
    start and end hour."""
    if sh == 0 and eh == 0:
        return data
    return [row for row in data if sh <= (row.time.hour + row.time.minute / 60) <= eh]


def check_sun_above_horizon(data, metadata):
    """Remove non-daylight hour entries."""

    def solar_altitude_check(row: WeaData):
        alt, _ = solar_angle(
            metadata.latitude,
            metadata.longitude,
            metadata.timezone,
            row.time.month,
            row.time.day,
            row.time.hour + row.time.minute / 60,
        )
        return alt > 0

    return [row for row in data if solar_altitude_check(row)]


def filter_data_with_zero_dni(data):
    """Filter out data entries with zero direct normal irradiance."""
    return [row for row in data if row.dni != 0]


def solar_minute(data: WeaData) -> int:
    # assuming never leap year
    mo_da = [0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334]
    jd = mo_da[data.time.month - 1] + data.time.day
    return 24 * 60 * (jd - 1) + int(data.time.hour * 60.0 + data.time.minute + 0.5)
